import logging
import shutil
from pathlib import Path
from typing import Type

import numpy as np

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.import_parsers.ImportParser import ImportParser
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper
from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.MotifPerformancePlotHelper import MotifPerformancePlotHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class MotifTestSetPerformance(EncodingReport):
    """
    This report can be used to show the performance of a learned set motifs using the :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder`
    on an independent test set of unseen data.

    It is recommended to first run the report :py:obj:`~immuneML.reports.data_reports.MotifGeneralizationAnalysis.MotifGeneralizationAnalysis`
    in order to calibrate the optimal recall thresholds and plot the performance of motifs on training- and validation sets.

    **Specification arguments:**

    - test_dataset (dict): parameters for importing a SequenceDataset to use as an independent test set. By default,
      the import parameters 'is_repertoire' and 'paired' will be set to False to ensure a SequenceDataset is imported.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_motif_report:
                    MotifTestSetPerformance:
                        test_dataset:
                            format: AIRR # choose any valid import format
                            params:
                                path: path/to/files/
                                is_repertoire: False  # is_repertoire must be False to import a SequenceDataset
                                paired: False         # paired must be False to import a SequenceDataset
                                # optional other parameters...

    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None,
                 test_dataset_import_cls: Type[DataImport] = None,
                 test_dataset_import_params: DatasetImportParams = None,
                 training_set_name: str = None, test_set_name: str = None,
                 split_by_motif_size: bool = None,
                 highlight_motifs_path: str = None, highlight_motifs_name: str = None,
                 min_points_in_window: int = None,
                 smoothing_constant1: float = None, smoothing_constant2: float = None,
                 keep_test_dataset: bool = None,
                 number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.test_dataset_import_cls = test_dataset_import_cls
        self.test_dataset_import_params = test_dataset_import_params

        self.keep_test_dataset = keep_test_dataset
        self.split_by_motif_size = split_by_motif_size
        self.training_set_name = training_set_name
        self.test_set_name = test_set_name
        self.highlight_motifs_path = highlight_motifs_path
        self.highlight_motifs_name = highlight_motifs_name
        self.min_points_in_window = min_points_in_window
        self.smoothing_constant1 = smoothing_constant1
        self.smoothing_constant2 = smoothing_constant2

    @classmethod
    def build_object(cls, **kwargs):
        location = MotifTestSetPerformance.__name__

        import_cls, test_dataset_import_params = MotifTestSetPerformance._parse_dataset_params(kwargs)

        kwargs["test_dataset_import_cls"] = import_cls
        kwargs["test_dataset_import_params"] = test_dataset_import_params

        del kwargs["test_dataset"]

        ParameterValidator.assert_type_and_value(kwargs["split_by_motif_size"], bool, location, "split_by_motif_size")
        ParameterValidator.assert_type_and_value(kwargs["training_set_name"], str, location, "training_set_name")
        ParameterValidator.assert_type_and_value(kwargs["test_set_name"], str, location, "test_set_name")
        ParameterValidator.assert_type_and_value(kwargs["keep_test_dataset"], bool, location, "keep_test_dataset")
        ParameterValidator.assert_type_and_value(kwargs["min_points_in_window"], int, location, "min_points_in_window",
                                                 min_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["smoothing_constant1"], (int, float), location,
                                                 "smoothing_constant1", min_exclusive=0)
        ParameterValidator.assert_type_and_value(kwargs["smoothing_constant2"], (int, float), location,
                                                 "smoothing_constant2", min_exclusive=0)

        if "highlight_motifs_path" in kwargs and kwargs["highlight_motifs_path"] is not None:
            PositionalMotifHelper.check_motif_filepath(kwargs["highlight_motifs_path"], location,
                                                       "highlight_motifs_path")

        ParameterValidator.assert_type_and_value(kwargs["highlight_motifs_name"], str, location,
                                                 "highlight_motifs_name")

        return MotifTestSetPerformance(**kwargs)

    @staticmethod
    def _parse_dataset_params(kwargs):
        location = MotifTestSetPerformance.__name__

        ParameterValidator.assert_type_and_value(kwargs["test_dataset"], dict, location, "test_dataset")
        ParameterValidator.assert_keys_present(kwargs["test_dataset"].keys(), ["format", "params"], location,
                                               "test_dataset")
        ParameterValidator.assert_type_and_value(kwargs["test_dataset"]["format"], str, location, "test_dataset/format")

        import_cls = ReflectionHandler.get_class_by_name("{}Import".format(kwargs["test_dataset"]["format"]))
        default_params = DefaultParamsLoader.load(ImportParser.keyword, kwargs["test_dataset"]["format"])
        params_dict = {**default_params, **kwargs["test_dataset"]["params"]}

        test_dataset_import_params = DatasetImportParams.build_object(**params_dict)

        if test_dataset_import_params.is_repertoire:
            logging.warning(f"{location}: This report only allows the reference dataset to be of type SequenceDataset. "
                          "Setting 'test_dataset/params/is_repertoire' to False...")
            test_dataset_import_params.is_repertoire = False

        if test_dataset_import_params.paired:
            logging.warning(f"{location}: This report only allows the reference dataset to be of type SequenceDataset. "
                          "Setting 'test_dataset/params/paired' to False...")
            test_dataset_import_params.paired = False

        return import_cls, test_dataset_import_params

    def _generate(self) -> ReportResult:
        test_dataset = self._get_test_dataset()
        test_encoded_data = self._encode_test_data(test_dataset)

        training_plotting_data, test_plotting_data = MotifPerformancePlotHelper.get_plotting_data(
            self.dataset.encoded_data,
            test_encoded_data.encoded_data,
            self.highlight_motifs_path,
            self.highlight_motifs_name)

        training_plotting_data["motif_size"] = training_plotting_data["feature_names"].apply(
            PositionalMotifHelper.get_motif_size)
        test_plotting_data["motif_size"] = test_plotting_data["feature_names"].apply(
            PositionalMotifHelper.get_motif_size)

        output_tables, output_plots = self._get_report_outputs(training_plotting_data, test_plotting_data)

        if not self.keep_test_dataset:
            shutil.rmtree(self.test_dataset_import_params.result_path)

        return ReportResult(name=self.name,
                            info="Performance of motifs on an independent test set",
                            output_figures=output_plots,
                            output_tables=output_tables)

    def _get_report_outputs(self, training_plotting_data, test_plotting_data):
        if self.split_by_motif_size:
            return self._construct_and_plot_data_per_motif_size(training_plotting_data, test_plotting_data)
        else:
            return self._construct_and_plot_data(training_plotting_data, test_plotting_data)

    def _construct_and_plot_data_per_motif_size(self, training_plotting_data, test_plotting_data):
        output_tables, output_plots = [], []

        for motif_size in sorted(set(training_plotting_data["motif_size"])):
            sub_training_plotting_data = training_plotting_data[training_plotting_data["motif_size"] == motif_size]
            sub_test_plotting_data = test_plotting_data[test_plotting_data["motif_size"] == motif_size]

            sub_output_tables, sub_output_plots = self._construct_and_plot_data(sub_training_plotting_data,
                                                                                sub_test_plotting_data,
                                                                                motif_size=motif_size)

            output_tables.extend(sub_output_tables)
            output_plots.extend(sub_output_plots)

        return output_tables, output_plots

    def _construct_and_plot_data(self, training_plotting_data, test_plotting_data, motif_size=None):
        training_combined_precision = self._get_combined_precision(training_plotting_data)
        test_combined_precision = self._get_combined_precision(test_plotting_data)

        motif_size_suffix = f"_motif_size={motif_size}" if motif_size is not None else ""
        motifs_name = f"motifs of length {motif_size}" if motif_size is not None else "motifs"

        output_tables = MotifPerformancePlotHelper.write_output_tables(self, training_plotting_data, test_plotting_data,
                                                                       training_combined_precision,
                                                                       test_combined_precision, motifs_name=motifs_name,
                                                                       file_suffix=motif_size_suffix)
        output_plots = MotifPerformancePlotHelper.write_plots(self, training_plotting_data, test_plotting_data,
                                                              training_combined_precision, test_combined_precision,
                                                              training_tp_cutoff="auto", test_tp_cutoff="auto",
                                                              motifs_name=motifs_name, file_suffix=motif_size_suffix)

        return output_tables, output_plots

    def _get_combined_precision(self, plotting_data):
        return MotifPerformancePlotHelper.get_combined_precision(plotting_data,
                                                                 min_points_in_window=self.min_points_in_window,
                                                                 smoothing_constant1=self.smoothing_constant1,
                                                                 smoothing_constant2=self.smoothing_constant2)

    def _write_output_tables(self, training_plotting_data, test_plotting_data, training_combined_precision,
                             test_combined_precision, file_suffix=""):
        results_table_name = "Confusion matrix and precision/recall scores for significant motifs on the {}"
        combined_precision_table_name = "Combined precision scores of motifs on the {} for each TP value on the " + str(
            self.training_set_name)

        train_results_table = self._write_output_table(training_plotting_data,
                                                       self.result_path / f"training_set_scores{file_suffix}.csv",
                                                       results_table_name.format(self.training_set_name))
        test_results_table = self._write_output_table(test_plotting_data,
                                                      self.result_path / f"test_set_scores{file_suffix}.csv",
                                                      results_table_name.format(self.test_set_name))
        training_combined_precision_table = self._write_output_table(training_combined_precision,
                                                                     self.result_path / f"training_combined_precision{file_suffix}.csv",
                                                                     combined_precision_table_name.format(
                                                                         self.training_set_name))
        test_combined_precision_table = self._write_output_table(test_combined_precision,
                                                                 self.result_path / f"test_combined_precision{file_suffix}.csv",
                                                                 combined_precision_table_name.format(
                                                                     self.test_set_name))

        return [table for table in [train_results_table, test_results_table, training_combined_precision_table,
                                    test_combined_precision_table] if table is not None]

    def _plot_precision_per_tp(self, file_path, plotting_data, combined_precision, dataset_type, tp_cutoff,
                               motifs_name="motifs"):
        return MotifPerformancePlotHelper.plot_precision_per_tp(file_path, plotting_data, combined_precision,
                                                                dataset_type,
                                                                training_set_name=self.training_set_name,
                                                                motifs_name=motifs_name,
                                                                tp_cutoff=tp_cutoff,
                                                                highlight_motifs_name=self.highlight_motifs_name)

    def _plot_precision_recall(self, file_path, plotting_data, min_recall=None, min_precision=None, dataset_type=None,
                               motifs_name="motifs"):
        return MotifPerformancePlotHelper.plot_precision_recall(file_path, plotting_data, min_recall=min_recall,
                                                                min_precision=min_precision,
                                                                dataset_type=dataset_type, motifs_name=motifs_name,
                                                                highlight_motifs_name=self.highlight_motifs_name)

    def _encode_test_data(self, test_dataset):
        encoder = self._get_encoder()
        params = EncoderParams(result_path=self.result_path / "encoded_test_dataset",
                               label_config=self._get_label_config(),
                               pool_size=self.number_of_processes,
                               learn_model=False)

        return encoder.encode(test_dataset, params)

    def _get_encoder(self):
        encoder = MotifEncoder(label=self._get_label_name(),
                               name=f"motif_encoder_{self.name}")

        encoder.learned_motif_filepath = self.dataset.encoded_data.info["learned_motif_filepath"]

        return encoder

    def _get_test_dataset_y_true(self, test_dataset):
        label_name = self._get_label_name()
        positive_class = self._get_positive_class()

        y_true = [sequence.get_attribute(label_name) == positive_class for sequence in test_dataset.get_data()]

        return np.array(y_true)

    def _get_motifs(self):
        motif_names = self._get_motif_names()
        return [PositionalMotifHelper.string_to_motif(name, "&", "-") for name in motif_names]

    def _get_motif_names(self):
        return list(self.dataset.encoded_data.feature_annotations.feature_names)

    def _get_test_dataset(self):
        self._set_result_path()
        test_dataset = self._import_test_dataset()
        self._check_test_dataset(test_dataset)

        return test_dataset

    def _check_test_dataset(self, test_dataset):
        self._check_sequence_length(test_dataset)
        self._check_dataset_label(test_dataset)

    def _check_sequence_length(self, test_dataset):
        legal_length = self._get_legal_sequence_length()

        for sequence in test_dataset.get_data():
            assert len(sequence.get_sequence()) == legal_length, \
                (f"{MotifTestSetPerformance.__name__}: the length of the sequences in the test dataset is required "
                 f"to match the length of the original dataset ({legal_length}). Found sequence of length: "
                 f"{len(sequence.get_sequence())}")

    def _get_legal_sequence_length(self):
        return self.dataset.data.cdr3_aa.lengths[0]

    def _check_dataset_label(self, test_dataset):
        label_name = self._get_label_name()
        label_values = set(self.dataset.encoded_data.labels[label_name])

        assert label_name in list(test_dataset.dynamic_fields.keys()), \
            f"{self.__class__.__name__}: Label {label_name} must be set for the test dataset."

        test_dataset_label_values = set(test_dataset.get_metadata([label_name])[label_name])

        assert label_values == test_dataset_label_values,  \
            (f"{self.__class__.__name__}: Label {label_name} with classes {', '.join(label_values)} must be set for "
             f"the test dataset. Instead, it has the following classes: {', '.join(test_dataset_label_values)}.")

    def _get_label_config(self):
        return LabelConfiguration([self._get_label()])

    def _get_label(self):
        label_name = self._get_label_name()
        label_values = list(set(self.dataset.encoded_data.labels[label_name]))
        positive_class = self._get_positive_class()

        return Label(name=label_name, values=label_values, positive_class=positive_class)

    def _get_label_name(self):
        return list(self.dataset.encoded_data.labels.keys())[0]

    def _get_positive_class(self):
        return self.dataset.encoded_data.info["positive_class"]

    def _set_result_path(self):
        self.test_dataset_import_params.result_path = self.result_path / f"test_dataset_{self.name}"

    def _import_test_dataset(self):
        return self.test_dataset_import_cls(self.test_dataset_import_params,
                                            f"test_dataset_{self.name}").import_dataset()

    def check_prerequisites(self) -> bool:
        location = MotifTestSetPerformance.__name__

        if self.dataset.encoded_data is None or self.dataset.encoded_data.info is None:
            logging.warning(f"{location}: the dataset is not encoded, skipping this report...")
            return False
        elif self.dataset.encoded_data.encoding != MotifEncoder.__name__:
            logging.warning(
                f"{location}: the dataset encoding ({self.dataset.encoded_data.encoding}) "
                f"does not match the required encoding ({MotifEncoder.__name__}), skipping this report...")
            return False
        elif self.dataset.encoded_data.feature_annotations is None:
            logging.warning(f"{location}: missing feature annotations for {MotifEncoder.__name__},"
                            f"skipping this report...")
            return False
        else:
            return True



import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.SequenceParams import ChainPair, RegionType
from immuneML.data_model.SequenceSet import Repertoire
from scripts.specification_util import update_docs_per_mapping


class TenxGenomicsImport(DataImport):
    """
    Imports data from the 10x Genomics Cell Ranger analysis pipeline into a Repertoire-, Sequence- or ReceptorDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.

    .. note::

        The 10xGenomics Cell Ranger VDJ pipeline also directly exports data in AIRR format ('airr_rearrangement.tsv',
        see https://www.10xgenomics.com/support/software/cell-ranger/latest/analysis/outputs/cr-5p-outputs-overview-vdj).
        If possible, we highly recommend directly using the AIRR formatted files as input for immuneML.


    If AIRR files are not available, this importer may be used to import data from Contig annotation CSV files
    ('all_contig_annotations.csv' or 'filtered_contig_annotations.csv') as described here: https://www.10xgenomics.com/support/software/cell-ranger/latest/analysis/outputs/cr-5p-outputs-annotations-vdj#contig-annotation-csv

    It is recommended to run :py:obj:`~immuneML.preprocessing.filters.DuplicateSequenceFilter.DuplicateSequenceFilter` to
    collapse together clonotypes when importing as a RepertoireDataset, and in some cases for Sequence- and ReceptorDatasets.

    Note that for pairing together Receptor chains, the column named 'barcode' is used.


    **Specification arguments:**

    - path (str): For RepertoireDatasets, this is the path to a directory with 10xGenomics files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions.For setting Sequence- or ReceptorDataset labels, metadata_file is ignored, use label_columns instead.

    - label_columns (list): For Sequence- or ReceptorDataset, this parameter can be used to explicitly set the column names of labels to import. These labels can be used as prediction target. When label_columns are not set, label names are attempted to be discovered automatically (any column name which is not used in the column_mapping). For setting RepertoireDataset labels, label_columns is ignored, use metadata_file instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the 10xGenomics column named 'clonotype_id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor.  Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_productive (bool): Whether productive sequences (with value 'True' in column productive) should be included in the imported sequences. By default, import_productive is True.

    - import_unproductive (bool): Whether productive sequences (with value 'Fale' in column productive) should be included in the imported sequences. By default, import_unproductive is False.

    - import_unknown_productivity (bool): Whether sequences with unknown productivity (missing or 'NA' value in column productive) should be included in the imported sequences. By default, import_unknown_productivity is True.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as 10xGenomics uses IMGT junction as CDR3. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from 10xGenomics column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the 10xGenomics file, or using alternative column names). Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet). For 10xGenomics, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

                cdr3: junction
                cdr3_nt: junction_aa
                v_gene: v_call
                j_gene: j_call
                umis: duplicate_count
                clonotype_id: cell_id
                consensus_id: sequence_id

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For 10xGenomics format, there is no default column_mapping_synonyms.

    - separator (str): Column separator, for 10xGenomics this is by default ",".


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_10x_dataset:
                    format: 10xGenomics
                    params:
                        path: path/to/files/
                        is_repertoire: True # whether to import a RepertoireDataset
                        metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                        paired: False # whether to import SequenceDataset (False) or ReceptorDataset (True) when is_repertoire = False
                        receptor_chains: TRA_TRB # what chain pair to import for a ReceptorDataset
                        import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                        import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                        import_empty_aa_sequences: False # filter out sequences if they don't have amino acid sequence set
                        # Optional fields with 10xGenomics-specific defaults, only change when different behavior is required:
                        separator: "," # column separator
                        region_type: IMGT_CDR3 # what part of the sequence to import
                        column_mapping: # column mapping 10xGenomics: immuneML
                            cdr3: junction_aa
                            cdr3_nt: junction
                            v_gene: v_call
                            j_gene: j_call
                            umis: duplicate_count
                            clonotype_id: cell_id
                            consensus_id: sequence_id

    """

    @staticmethod
    def get_documentation():
        doc = str(TenxGenomicsImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet)": f"Valid immuneML fields that can be specified here by `the AIRR Rearrangement Schema <https://docs.airr-community.org/en/latest/datarep/rearrangements.html>`_."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc





import shutil

import argparse
import sys
from pathlib import Path

from immuneML.data_model.bnp_util import write_yaml, read_yaml
from immuneML.util.PathBuilder import PathBuilder


def parse_command_line_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building specification for applying trained generative "
                                                 "model in Galaxy")

    parser.add_argument("-m", "--ml_config_path", required=True,
                        help="Path to the exported ml_config zip file with the trained model.")
    parser.add_argument("-e", "--gen_example_count", required=True,
                        help="Number of examples to generate.")
    parser.add_argument("-s", "--sequence_length_report", choices=["True", "False"], default="False",
                        help="Whether to run the SequenceLengthDistribution report.")
    parser.add_argument("-q", "--amino_acid_report", choices=["True", "False"], default="False",
                        help="Whether to run the AminoAcidFrequencyDistribution report.")
    parser.add_argument("-g", "--gen_model_overview", choices=["True", "False"], default="False",
                        help="Whether to run the model report.")
    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def build_specs(parsed_args):
    reports = {}
    if parsed_args.sequence_length_report == "True":
        reports['sequence_length_distribution'] = "SequenceLengthDistribution"
    if parsed_args.amino_acid_report == "True":
        reports['amino_acid_frequency_distribution'] = "AminoAcidFrequencyDistribution"

    model_overview_reports = {
        "PWM": "PWMSummary",
        "SimpleVAE": "VAESummary"
    }

    if parsed_args.gen_model_overview == "True":
        shutil.unpack_archive(parsed_args.ml_config_path, "./unpacked", 'zip')
        model_overview = read_yaml(Path('unpacked/model_overview.yaml'))
        if model_overview['type'] in model_overview_reports:
            reports['generative_model_overview'] = model_overview_reports[model_overview['type']]
        shutil.rmtree("./unpacked")

    specs = {
        "definitions": {
            "reports": reports
        },
        "instructions": {
            "inst1": {
                "type": "ApplyGenModel",
                "gen_examples_count": int(parsed_args.gen_example_count),
                "reports": list(reports.keys()),
                "ml_config_path": parsed_args.ml_config_path,
            }
        }
    }

    return specs


def main(args):
    parsed_args = parse_command_line_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    write_yaml(output_location, specs)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



import dataclasses
from typing import Protocol
from scipy.stats import poisson
import numpy as np
from bionumpy import EncodedRaggedArray, EncodedArray, count_encoded

from immuneML.ml_methods.generative_models.SequenceTransitionDistribution import EncodedLookup


@dataclasses.dataclass
class Poisson:
    mu: float

    def log_prob(self, x):
        return poisson.logpmf(x, self.mu)

    def sample(self, n_samples):
        return poisson.rvs(self.mu, size=n_samples)


class KmerDistribution(Protocol):
    def sample(self, count: int) -> EncodedRaggedArray:
        ...

    def log_prob(self, kmers) -> np.ndarray:
        ...


class EmpiricalLengthDistribution:
    def __init__(self, lengths_frequencies: np.ndarray):
        self.lengths_frequencies = lengths_frequencies

    def sample(self, count: int) -> np.ndarray:
        return np.random.choice(np.arange(len(self.lengths_frequencies)), size=count, p=self.lengths_frequencies)

    def log_prob(self, lengths: np.ndarray) -> np.ndarray:
        mask = lengths < len(self.lengths_frequencies)
        probs = np.zeros_like(lengths, dtype=float)
        probs[mask] = self.lengths_frequencies[lengths[mask]]
        return np.log(probs)


class SmoothedLengthDistribution:
    def __init__(self, empirical_distribution, smooth_distribution, p_smooth):
        self.empirical_distribution = empirical_distribution
        self.smooth_distribution = smooth_distribution
        self.p_smooth = p_smooth

    def sample(self, count: int) -> np.ndarray:
        is_smooth = np.random.rand(count) < self.p_smooth
        smooth_samples = self.smooth_distribution.sample(count)
        empirical_samples = self.empirical_distribution.sample(count)
        return np.where(is_smooth, smooth_samples, empirical_samples)

    def log_prob(self, lengths: np.ndarray) -> np.ndarray:
        return np.logaddexp(np.log(1 - self.p_smooth) + self.empirical_distribution.log_prob(lengths),
                            np.log(self.p_smooth) + self.smooth_distribution.log_prob(lengths))


class KmerModel:
    def __init__(self, kmer_probs: EncodedLookup):
        self.kmer_probs = kmer_probs
        self._raw_values = np.arange(kmer_probs.alphabet_size, dtype=np.uint8)
        self._log_probs = np.log(kmer_probs)

    def sample(self, count: int) -> EncodedRaggedArray:
        kmer_hashes = np.random.choice(self._raw_values, size=count, p=self.kmer_probs.raw())
        return EncodedArray(kmer_hashes, self.kmer_probs.encoding)

    def log_prob(self, kmers: EncodedRaggedArray) -> np.ndarray:
        return self._log_probs[kmers]


class MultinomialKmerModel:
    def __init__(self, kmer_probs: EncodedLookup, sequence_length: int):
        self.kmer_model = KmerModel(kmer_probs)
        self.sequence_length = sequence_length

    def sample(self, count: int) -> EncodedRaggedArray:
        if hasattr(self.sequence_length, 'sample'):
            sequence_length = self.sequence_length.sample(count)
        else:
            sequence_length = np.full((count,), self.sequence_length, dtype=int)
        total_kmers = sequence_length.sum()
        kmers = self.kmer_model.sample(total_kmers)
        # kmer_hashes = np.random.choice(self._raw_values, size=total_kmers,
        # p=self.kmer_probs.raw())
        # kmers = EncodedArray(kmer_hashes, self.kmer_probs.encoding)
        return EncodedRaggedArray(kmers, sequence_length)

    def log_prob(self, kmers: EncodedRaggedArray) -> np.ndarray:
        lengths = kmers.shape[-1]
        if hasattr(self.sequence_length, 'log_prob'):
            length_log_probs = self.sequence_length.log_prob(lengths)
        else:
            assert np.all(lengths == self.sequence_length)
            length_log_probs = 0
        return self.kmer_model.log_prob(kmers).sum(axis=-1) + length_log_probs


def estimate_length_distribution(lengths: np.ndarray) -> EmpiricalLengthDistribution:
    counts = np.bincount(lengths)
    return EmpiricalLengthDistribution(counts / counts.sum())


def estimate_smoothed_length_distribution(lengths: np.ndarray, prior_count=1) -> SmoothedLengthDistribution:
    empirical = estimate_length_distribution(lengths)
    smooth = Poisson(np.mean(lengths))
    return SmoothedLengthDistribution(empirical, smooth, prior_count / (prior_count + len(lengths)))


def estimate_kmer_model(kmers: EncodedRaggedArray, prior_count=1) -> MultinomialKmerModel:
    length_distribution = estimate_smoothed_length_distribution(kmers.lengths, prior_count=prior_count)
    kmer_counts = count_encoded(kmers, axis=None)
    lookup = EncodedLookup((kmer_counts.counts + prior_count / len(kmer_counts.counts)) / (kmers.size + prior_count),
                           kmers.encoding)

    return MultinomialKmerModel(lookup, length_distribution)



import logging
import os
from pathlib import Path

from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.ImportHelper import ImportHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler


class MatchedReferenceUtil:
    """
    Utility class for MatchedSequencesEncoder and MatchedReceptorsEncoder
    """

    @staticmethod
    def prepare_reference(reference_params: dict, location: str, paired: bool):
        ParameterValidator.assert_keys(list(reference_params.keys()), ["format", "params"], location, "reference")

        seq_import_params = reference_params["params"] if "params" in reference_params else {}

        assert os.path.isfile(seq_import_params["path"]), f"{location}: the file {seq_import_params['path']} does not exist. " \
                                                          f"Specify the correct path under reference."

        if "is_repertoire" in seq_import_params:
            assert seq_import_params["is_repertoire"] is False, f"{location}: is_repertoire must be False for SequenceImport"
        else:
            seq_import_params["is_repertoire"] = False

        if "paired" in seq_import_params:
            assert seq_import_params["paired"] == paired, f"{location}: paired must be {paired} for SequenceImport"
        else:
            seq_import_params["paired"] = paired

        format_str = reference_params["format"]

        import_class = ReflectionHandler.get_class_by_name(f"{format_str}Import")
        assert import_class is not None, (f"{MatchedReferenceUtil.__name__}: {format_str} could not be imported. "
                                          f"Check if the format name has been written correctly.")
        default_params = DefaultParamsLoader.load(EnvironmentSettings.default_params_path / "datasets",
                                                  DefaultParamsLoader.convert_to_snake_case(format_str))

        params = {**default_params, **seq_import_params}

        path = Path(reference_params['params']['path'])
        params['result_path'] = PathBuilder.build(path.parent / 'iml_imported' if path.is_file() else path / 'iml_imported')

        if format_str == "SingleLineReceptor":
            receptors = list(import_class(params, 'tmp_receptor_dataset').import_dataset().get_data())
        else:
            receptors = list(import_class(params=params, dataset_name="tmp_dataset").import_dataset().get_data())

        assert len(receptors) > 0, f"MatchedReferenceUtil: The total number of imported reference {'receptors' if paired else 'sequences'} is 0, please ensure that reference import is specified correctly."
        logging.info(f"MatchedReferenceUtil: successfully imported {len(receptors)} reference {'receptors' if paired else 'sequences'}.")

        return receptors



import copy
import logging
from pathlib import Path
from typing import List

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.Report import Report
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.ParameterValidator import ParameterValidator


class ReportUtil:

    @staticmethod
    def _make_new_report(report: Report, path: Path, number_of_processes: int = 1, context: dict = None):
        tmp_report = copy.deepcopy(report)
        report_name = report.name if report.name is not None else 'report_result'
        tmp_report.result_path = path / report_name
        tmp_report.number_of_processes = number_of_processes
        tmp_report.set_context(context)
        return tmp_report

    @staticmethod
    def run_ML_reports(train_dataset: Dataset, test_dataset: Dataset, method: MLMethod, reports: List[MLReport], path: Path,
                       hp_setting: HPSetting, label: Label, number_of_processes: int = 1, context: dict = None) -> List[ReportResult]:
        report_results = []
        for report in reports:
            tmp_report = ReportUtil._make_new_report(report, path, number_of_processes, context)
            tmp_report.method = method
            tmp_report.train_dataset = train_dataset
            tmp_report.test_dataset = test_dataset
            tmp_report.hp_setting = hp_setting
            tmp_report.label = label
            result = tmp_report.generate_report()
            report_results.append(result)
        return report_results

    @staticmethod
    def _run_reports_on_dataset(dataset: Dataset, reports: list, path: Path, number_of_processes: int = 1, context: dict = None) -> List[ReportResult]:
        report_results = []
        for report in reports:
            tmp_report = ReportUtil._make_new_report(report, path, number_of_processes, context)
            tmp_report.dataset = dataset
            result = tmp_report.generate_report()
            report_results.append(result)
        return report_results

    @staticmethod
    def run_encoding_reports(dataset: Dataset, reports: List[EncodingReport], path: Path, number_of_processes: int = 1, context: dict = None) -> List[ReportResult]:
        return ReportUtil._run_reports_on_dataset(dataset, reports, path, number_of_processes, context)

    @staticmethod
    def run_data_reports(dataset: Dataset, reports: List[DataReport], path: Path, number_of_processes: int = 1, context: dict = None):
        return ReportUtil._run_reports_on_dataset(dataset, reports, path, number_of_processes, context)

    @staticmethod
    def update_split_by_label_kwargs(kwargs, location):
        if kwargs["label"] is not None:
            ParameterValidator.assert_type_and_value(kwargs["label"], str, location, "label")

            if kwargs["split_by_label"] is False:
                logging.warning(f"{location}: label is set but split_by_label was False, setting split_by_label to True")
                kwargs["split_by_label"] = True


import io
import os
import statistics
from pathlib import Path

import pandas as pd

from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.hyperparameter_optimization.states.HPAssessmentState import HPAssessmentState
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.hyperparameter_optimization.states.HPLabelState import HPLabelState
from immuneML.hyperparameter_optimization.states.HPSelectionState import HPSelectionState
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.reports.ReportResult import ReportResult
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.StringHelper import StringHelper


class HPHTMLBuilder:
    """
    A class that will make HTML file(s) out of TrainMLModelState object to show what analysis took place in the TrainMLModel.
    """

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"
    NUM_DIGITS = 3

    @staticmethod
    def build(state: TrainMLModelState = None) -> Path:
        """
        Function that builds the HTML files based on the HPOptimization state.
        Arguments:
            state: HPOptimizationState object with all details on the optimization
        Returns:
             path to the main HTML file (index.html which is located under state.result_path)
        """

        base_path = PathBuilder.build(state.path / "../HTML_output/")
        state = HPHTMLBuilder._move_reports_recursive(state, base_path)
        html_map = HPHTMLBuilder._make_main_html_map(state, base_path)
        result_file = base_path / f"TrainMLModelReport_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "HPOptimization.html",
                             template_map=html_map, result_path=result_file)

        for label_name in state.label_configuration.get_labels_by_name():
            for index, item in enumerate(HPHTMLBuilder._make_assessment_pages(state, base_path, label_name)):
                TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "AssessmentSplitDetails.html",
                                     template_map=item,
                                     result_path=base_path / HPHTMLBuilder._make_assessment_split_path(index, state.name, label_name))

        for label_name in state.label_configuration.get_labels_by_name():
            for assessment_index in range(state.assessment.split_count):
                TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "SelectionDetails.html",
                                     template_map=HPHTMLBuilder._make_selection(state, assessment_index, label_name, base_path),
                                     result_path=base_path / HPHTMLBuilder._make_selection_split_path(assessment_index, label_name, state.name))

        return result_file

    @staticmethod
    def _make_assessment_split_path(split_index: int, state_name: str, label_name: str) -> Path:
        return Path(f"{state_name}_CV_assessment_split_{split_index + 1}_{label_name}.html")

    @staticmethod
    def _make_selection_split_path(assessment_index: int, label_name: str, state_name: str) -> Path:
        return Path(f"{state_name}_selection_details_{label_name}_split_{assessment_index + 1}.html")

    @staticmethod
    def _make_selection(state: TrainMLModelState, assessment_index: int, label_name: str, base_path):
        selection_state = state.assessment_states[assessment_index].label_states[label_name].selection_state

        hp_settings = []
        optimal = selection_state.optimal_hp_setting.get_key()

        for hp_setting, hp_items in selection_state.hp_items.items():
            hp_splits = []
            for hp_item in hp_items:
                hp_splits.append(HPHTMLBuilder._print_metric(hp_item.performance, state.optimization_metric))
            hp_settings.append({
                "hp_setting": hp_setting,
                "hp_splits": hp_splits,
                "optimal": hp_setting == optimal
            })

            if len(hp_splits) > 1:
                hp_settings[-1]["average"] = round(statistics.mean(perf for perf in hp_splits if [isinstance(perf, float)]), HPHTMLBuilder.NUM_DIGITS)
                hp_settings[-1]["show_average"] = True
            else:
                hp_settings[-1]["average"] = None
                hp_settings[-1]["show_average"] = False

            hp_settings[-1]["hp_splits"] = [{"optimization_metric_val": val} for val in hp_settings[-1]["hp_splits"]]

        has_other_metrics = len([metric for metric in state.metrics if metric != state.optimization_metric]) > 0 and \
                            not (state.selection.split_strategy == SplitType.RANDOM and state.selection.training_percentage == 1)

        return {
            "css_style": Util.get_css_content(HPHTMLBuilder.CSS_PATH),
            "label": label_name,
            "assessment_split": assessment_index + 1,
            "splits": [{"split_index": i} for i in range(1, state.selection.split_count + 1)],
            "split_count": state.selection.split_count,
            "optimization_metric": state.optimization_metric.name.lower(),
            "has_other_metrics": has_other_metrics,
            "metrics": [{"performance": HPHTMLBuilder._extract_selection_performance_per_metric(selection_state, metric, state.selection.split_count),
                         "metric": HPHTMLBuilder._get_heading_metric_name(metric.name.lower())}
                        for metric in state.metrics if metric != state.optimization_metric] if has_other_metrics else None,
            "hp_settings": hp_settings,
            "show_average": any(hps["show_average"] for hps in hp_settings),
            "data_split_reports": [
                {'split_index': index + 1,
                 'train': Util.to_dict_recursive(selection_state.train_data_reports[index], base_path)
                 if len(selection_state.train_data_reports) == state.selection.split_count else None,
                 'test': Util.to_dict_recursive(selection_state.val_data_reports[index], base_path)
                 if len(selection_state.train_data_reports) == state.selection.split_count else None}
                for index in range(state.selection.split_count)] if len(state.selection.reports.data_split_reports) > 0 else None,
            "has_data_split_reports": len(state.selection.reports.data_split_reports) > 0,
            "has_reports_per_setting": len(state.selection.reports.encoding_reports) + len(state.selection.reports.model_reports) > 0,
            "reports_per_setting": [{
                "hp_setting": hp_setting,
                "reports": HPHTMLBuilder._make_selection_reports_for_item_list(hp_items, base_path)
            } for hp_setting, hp_items in selection_state.hp_items.items()]
        }

    @staticmethod
    def _make_selection_reports_for_item_list(hp_items: list, base_path) -> list:
        result = []

        for split_index, hp_item in enumerate(hp_items):
            result.append({
                "split_index": split_index + 1,
                "has_encoding_train_reports": len(hp_item.encoding_train_results) > 0,
                "has_encoding_test_reports": len(hp_item.encoding_test_results) > 0,
                "has_ml_reports": len(hp_item.model_report_results) > 0,
                "encoding_train_reports": Util.to_dict_recursive(hp_item.encoding_train_results, base_path) if len(
                    hp_item.encoding_train_results) > 0 else None,
                "encoding_test_reports": Util.to_dict_recursive(hp_item.encoding_test_results, base_path) if len(
                    hp_item.encoding_test_results) > 0 else None,
                "ml_reports": Util.to_dict_recursive(hp_item.model_report_results, base_path) if len(hp_item.model_report_results) > 0 else None,
            })

        return result if len(result) > 0 else None

    @staticmethod
    def _make_assessment_pages(state: TrainMLModelState, base_path: Path, label_name: str):
        assessment_list = []

        for i, assessment_state in enumerate(state.assessment_states):

            assessment_item = {"css_style": Util.get_css_content(HPHTMLBuilder.CSS_PATH),
                               "optimization_metric": state.optimization_metric.name.lower(),
                               "split_index": assessment_state.split_index + 1,
                               "hp_settings": [],
                               "has_reports": len(state.assessment.reports.model_reports) + len(state.assessment.reports.encoding_reports) > 0,
                               "train_data_reports": Util.to_dict_recursive(assessment_state.train_val_data_reports, base_path),
                               "test_data_reports": Util.to_dict_recursive(assessment_state.test_data_reports, base_path),
                               "show_data_reports": len(assessment_state.train_val_data_reports) > 0 or len(assessment_state.test_data_reports) > 0}

            if hasattr(assessment_state.train_val_dataset, "metadata_file") and assessment_state.train_val_dataset.metadata_file is not None:
                assessment_item["train_metadata_path"] = os.path.relpath(str(assessment_state.train_val_dataset.metadata_file), str(base_path))
                assessment_item["train_metadata"] = Util.get_table_string_from_csv(assessment_state.train_val_dataset.metadata_file)
            else:
                assessment_item["train_metadata_path"] = None

            if hasattr(assessment_state.test_dataset, "metadata_file") and assessment_state.test_dataset.metadata_file is not None:
                assessment_item['test_metadata_path'] = os.path.relpath(assessment_state.test_dataset.metadata_file, base_path)
                assessment_item["test_metadata"] = Util.get_table_string_from_csv(assessment_state.test_dataset.metadata_file)
            else:
                assessment_item["test_metadata_path"] = None

            assessment_item["label"] = label_name
            for hp_setting, item in assessment_state.label_states[label_name].assessment_items.items():
                optimal = str(assessment_state.label_states[label_name].optimal_hp_setting.get_key())
                reports_path = HPHTMLBuilder._make_assessment_reports(state, i, hp_setting, assessment_state, label_name, base_path)
                assessment_item["hp_settings"].append({
                    "optimal": str(hp_setting) == optimal,
                    "hp_setting": str(hp_setting),
                    "optimization_metric_val": HPHTMLBuilder._print_metric(item.performance, state.optimization_metric),
                    "reports_path": reports_path
                })
            assessment_item["show_non_optimal"] = len(assessment_item["hp_settings"]) > 1

            assessment_item["selection_path"] = HPHTMLBuilder._make_selection_split_path(i, label_name, state.name)
            assessment_item['performances_per_metric'] = HPHTMLBuilder._extract_assessment_performances_per_metric(state, assessment_state, label_name)

            assessment_list.append(assessment_item)

        return assessment_list

    @staticmethod
    def _extract_assessment_performances_per_metric(state: TrainMLModelState, assessment_state: HPAssessmentState, label_name: str) -> str:
        performance_metric = {"setting": [], **{metric.name.lower(): [] for metric in state.metrics}}
        for hp_setting, hp_item in assessment_state.label_states[label_name].assessment_items.items():
            performance_metric['setting'].append(str(hp_setting))
            for metric in sorted(state.metrics, key=lambda metric: metric.name.lower()):
                performance_metric[metric.name.lower()].append(HPHTMLBuilder._print_metric(hp_item.performance, metric))

        s = io.StringIO()
        pd.DataFrame(performance_metric).rename(columns={"setting": 'Hyperparameter settings (preprocessing, encoding, ML method)'})\
            .to_csv(s, sep="\t", index=False)
        return Util.get_table_string_from_csv_string(s.getvalue(), separator="\t")

    @staticmethod
    def _make_assessment_reports(state, i, hp_setting_key, assessment_state, label_name: str, base_path: Path):
        path = base_path / f"{state.name}_{label_name}_{hp_setting_key}_assessment_reports_split_{i + 1}.html"

        hp_item = assessment_state.label_states[label_name].assessment_items[hp_setting_key]
        data = {
            "split_index": i + 1,
            "hp_setting": hp_setting_key,
            "label": label_name,
            "css_style": Util.get_css_content(HPHTMLBuilder.CSS_PATH),
            "has_encoding_reports": len(hp_item.encoding_train_results) > 0 or len(hp_item.encoding_test_results) > 0,
            "has_ml_reports": len(hp_item.model_report_results) > 0,
            "encoding_train_reports": Util.to_dict_recursive(hp_item.encoding_train_results, base_path) if len(
                hp_item.encoding_train_results) > 0 else None,
            "encoding_test_reports": Util.to_dict_recursive(hp_item.encoding_test_results, base_path) if len(
                hp_item.encoding_test_results) > 0 else None,
            "ml_reports": Util.to_dict_recursive(hp_item.model_report_results, base_path) if len(
                hp_item.model_report_results) > 0 else None
        }

        if data["has_ml_reports"] or data["has_encoding_reports"]:
            TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "Reports.html", template_map=data, result_path=path)
            return path.name
        else:
            return None

    @staticmethod
    def _make_hp_per_label(state: TrainMLModelState):
        mapping = []

        for label_name in state.label_configuration.get_labels_by_name():
            results = []
            for i, assessment_state in enumerate(state.assessment_states):
                results.append({
                    "index": assessment_state.split_index + 1,
                    "hp_setting": assessment_state.label_states[label_name].optimal_assessment_item.hp_setting,
                    "optimization_metric_val": HPHTMLBuilder._print_metric(assessment_state.label_states[label_name].optimal_assessment_item.performance,
                                                                           state.optimization_metric),
                    "split_details_path": HPHTMLBuilder._make_assessment_split_path(assessment_state.split_index, state.name, label_name)
                })

            mapping.append({"label": label_name, "assessment_results": results})

        return mapping

    @staticmethod
    def _print_metric(performance: dict, metric: ClassificationMetric):
        if performance is not None and metric.name.lower() in performance:
            if isinstance(performance[metric.name.lower()], float):
                return round(performance[metric.name.lower()], HPHTMLBuilder.NUM_DIGITS)
            else:
                return performance[metric.name.lower()]
        else:
            return Constants.NOT_COMPUTED

    @staticmethod
    def _make_model_per_label(state: TrainMLModelState, base_path: Path) -> list:

        mapping = []

        for label_name in state.label_configuration.get_labels_by_name():
            mapping.append({
                "label": label_name,
                "model_path": Path(os.path.relpath(path=str(state.optimal_hp_item_paths[label_name]), start=str(base_path)))
            })

        return mapping

    @staticmethod
    def _process_report_tables(report_result, base_path: Path) -> list:
        """Process table outputs from a report result into a format suitable for HTML display"""
        tables = []
        if hasattr(report_result, "output_tables") and report_result.output_tables:
            for table in report_result.output_tables:
                table_path = Path(os.path.relpath(path=str(table.path), start=str(base_path)))
                table_data = {
                    "name": table.name if hasattr(table, "name") else None,
                    "path": str(table_path),
                    "table": Util.get_table_string_from_csv(table.path) if table.path.suffix == '.csv' else None
                }
                tables.append(table_data)
        return tables

    @staticmethod
    def _make_main_html_map(state: TrainMLModelState, base_path: Path) -> dict:
        html_map = {**Util.make_dataset_html_map(state.dataset), **{
            "css_style": Util.get_css_content(HPHTMLBuilder.CSS_PATH),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            "labels": [{"name": label.name, "values": str(label.values)[1:-1]} for label in state.label_configuration.get_label_objects()],
            "optimization_metric": state.optimization_metric.name.lower(),
            "other_metrics": str([metric.name.lower() for metric in state.metrics])[1:-1].replace("'", ""),
            "metrics": [{"name": metric.name.lower()} for metric in state.metrics],
            "assessment_desc": state.assessment,
            "selection_desc": state.selection,
            "show_hp_reports": bool(state.report_results),
            'hp_reports': [{
                "name": report.name,
                "info": report.info if hasattr(report, "info") else None,
                "output_figures": Util.to_dict_recursive(report.output_figures, base_path) if hasattr(report, "output_figures") else None,
                "output_tables": HPHTMLBuilder._process_report_tables(report, base_path) if hasattr(report, "output_tables") else None
            } for report in state.report_results] if state.report_results else None,
            "hp_per_label": HPHTMLBuilder._make_hp_per_label(state),
            'models_per_label': HPHTMLBuilder._make_model_per_label(state, base_path),
            'immuneML_version': MLUtil.get_immuneML_version()
        }}

        return html_map

    @staticmethod
    def _move_reports_recursive(obj, path: Path):
        for attribute in (vars(obj) if not isinstance(obj, dict) else obj):
            attribute_value = getattr(obj, attribute) if not isinstance(obj, dict) else obj[attribute]
            if isinstance(attribute_value, list) and all(isinstance(item, ReportResult) for item in attribute_value):
                new_attribute_values = []
                for report_result in attribute_value:
                    new_attribute_values.append(Util.update_report_paths(report_result, path))
                setattr(obj, attribute, new_attribute_values)
            elif isinstance(attribute_value, list) and all(isinstance(item, HPAssessmentState) for item in attribute_value):
                obj = HPHTMLBuilder._process_list_recursively(obj, attribute, attribute_value, path)
            elif isinstance(attribute_value, dict) and all(
                    isinstance(item, HPLabelState) or isinstance(item, HPItem) for item in attribute_value.values()):
                obj = HPHTMLBuilder._process_dict_recursive(obj, attribute, attribute_value, path)
            elif isinstance(attribute_value, dict) and all(isinstance(item, list) for item in attribute_value.values()) and all(
                    all(isinstance(item, HPItem) for item in item_list) for item_list in attribute_value.values()):
                obj = HPHTMLBuilder._process_hp_items(obj, attribute, attribute_value, path)
            elif isinstance(attribute_value, HPSelectionState):
                setattr(obj, attribute, HPHTMLBuilder._move_reports_recursive(attribute_value, path))

        return obj

    @staticmethod
    def _process_hp_items(obj, attribute, attribute_value, path: Path):
        new_attribute_value = {}
        for hp_setting, hp_item_list in attribute_value.items():
            new_hp_item_list = []
            for hp_item in hp_item_list:
                new_hp_item_list.append(HPHTMLBuilder._move_reports_recursive(hp_item, path))
            new_attribute_value[hp_setting] = new_hp_item_list

        setattr(obj, attribute, new_attribute_value)

        return obj

    @staticmethod
    def _process_dict_recursive(obj, attribute, attribute_value, path: Path):
        for key, value in attribute_value.items():
            attribute_value[key] = HPHTMLBuilder._move_reports_recursive(value, path)
        setattr(obj, attribute, attribute_value)
        return obj

    @staticmethod
    def _process_list_recursively(obj, attribute, attribute_value, path: Path):
        new_attribute_values = []
        for item in attribute_value:
            new_attribute_values.append(HPHTMLBuilder._move_reports_recursive(item, path))
        setattr(obj, attribute, new_attribute_values)
        return obj

    @staticmethod
    def _extract_selection_performance_per_metric(selection_state: HPSelectionState, metric: ClassificationMetric, split_count):
        performance = {"setting": [], **{f"split {i + 1}": [] for i in range(split_count)}}
        for hp_setting, hp_item_list in selection_state.hp_items.items():
            performance['setting'].append(str(hp_setting))
            for index, hp_item in enumerate(hp_item_list):
                performance[f'split {index + 1}'].append(HPHTMLBuilder._print_metric(hp_item.performance, metric))

        s = io.StringIO()
        pd.DataFrame(performance).rename(columns={"setting": 'Hyperparameter settings (preprocessing, encoding, ML method)'}).to_csv(s, sep="\t",
                                                                                                                                     index=False)
        return Util.get_table_string_from_csv_string(s.getvalue(), separator="\t")

    @staticmethod
    def _get_heading_metric_name(metric: str):
        if metric != "auc":
            return " ".join(metric.split("_")).title()
        else:
            return metric.upper()



import copy
import logging
from dataclasses import field, dataclass
from pathlib import Path
from typing import Dict, List

import numpy as np

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.bnp_util import merge_dataclass_objects, bnp_write_to_file, write_dataset_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.reports.train_gen_model_reports.TrainGenModelReport import TrainGenModelReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.GenModelInstruction import GenModelState, GenModelInstruction
from immuneML.workflows.steps.data_splitter.DataSplitter import DataSplitter
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams


@dataclass
class TrainGenModelState:
    result_path: Path = None
    name: str = None
    gen_examples_count: int = None
    model_path: Path = None
    generated_dataset: Dataset = None
    exported_datasets: Dict[str, Path] = field(default_factory=dict)
    report_results: Dict[str, List[ReportResult]] = field(
        default_factory=lambda: {'data_reports': [], 'ml_reports': [], 'instruction_reports': []})
    combined_dataset: Dataset = None
    train_dataset: Dataset = None
    test_dataset: Dataset = None
    training_percentage: float = None


class TrainGenModelInstruction(GenModelInstruction):
    """

    TrainGenModel instruction implements training generative AIRR models on receptor level. Models that can be trained
    for sequence generation are listed under Generative Models section.

    This instruction takes a dataset as input which will be used to train a model, the model itself, and the number of
    sequences to generate to illustrate the applicability of the model. It can also produce reports of the fitted model
    and reports of original and generated sequences.

    To use the generative model previously trained with immuneML, see :ref:`ApplyGenModel` instruction.


    **Specification arguments:**

    - dataset: dataset to use for fitting the generative model; it has to be defined under definitions/datasets

    - method: which model to fit (defined previously under definitions/ml_methods)

    - number_of_processes (int): how many processes to use for fitting the model

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the fitted model

    - reports (list): list of report ids (defined under definitions/reports) to apply after fitting a generative model
      and generating gen_examples_count examples; these can be data reports (to be run on generated examples), ML
      reports (to be run on the fitted model)

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_train_gen_model_inst: # user-defined instruction name
                type: TrainGenModel
                dataset: d1 # defined previously under definitions/datasets
                method: model1 # defined previously under definitions/ml_methods
                gen_examples_count: 100
                number_of_processes: 4
                training_percentage: 0.7
                export_generated_dataset: True
                export_combined_dataset: False
                reports: [data_rep1, ml_rep2]

    """

    MAX_ELEMENT_COUNT_TO_SHOW = 10

    def __init__(self, dataset: Dataset = None, method: GenerativeModel = None,
                 number_of_processes: int = 1, gen_examples_count: int = 100, result_path: Path = None,
                 name: str = None, reports: list = None, export_generated_dataset: bool = True,
                 export_combined_dataset: bool = False, training_percentage: float = None):
        super().__init__(TrainGenModelState(result_path, name, gen_examples_count), method, reports)
        self.dataset = dataset
        self.number_of_processes = number_of_processes
        self.export_generated_dataset = export_generated_dataset
        self.export_combined_dataset = export_combined_dataset
        self.state.training_percentage = training_percentage

    def run(self, result_path: Path) -> TrainGenModelState:
        self._set_path(result_path)
        self._split_dataset()
        self._fit_model()
        self._save_model()
        self._gen_data()
        if self.export_generated_dataset:
            self._export_generated_dataset()
        self._make_and_export_combined_dataset()
        self._run_reports()

        return self.state

    def _split_dataset(self):
        if self.state.training_percentage != 1:
            split_params = DataSplitterParams(dataset=self.dataset, split_strategy=SplitType.RANDOM, split_count=1,
                                              training_percentage=self.state.training_percentage,
                                              paths=[self.state.result_path])
            train_datasets, test_datasets = DataSplitter.run(split_params)
            self.state.train_dataset = train_datasets[0]
            self.state.test_dataset = test_datasets[0]
        else:
            logging.info(f"{TrainGenModelInstruction.__name__}: training_percentage was set to 1 meaning that the full "
                         f"dataset will be used for fitting the generative model. All resulting comparison reports "
                         f"will then use the full original dataset as opposed to independent test dataset if the "
                         f"training percentage was less than 1.")
            self.state.train_dataset = self.dataset
            self.state.test_dataset = self.dataset

    def _fit_model(self):
        print_log(f"{self.state.name}: starting to fit the model", True)
        self.method.fit(self.state.train_dataset, self.state.result_path)
        print_log(f"{self.state.name}: fitted the model", True)

    def _make_combined_dataset(self):
        path = PathBuilder.build(self.state.result_path / 'combined_dataset')

        gen_data = self._get_dataclass_object_from_dataset(self.generated_dataset,
                                                           np.ones(self.state.gen_examples_count),
                                                           np.zeros(self.state.gen_examples_count))

        if self.state.training_percentage < 1:

            org_data = self._get_dataclass_object_from_dataset(self.state.train_dataset,
                                                               np.zeros(self.state.train_dataset.get_example_count()),
                                                               np.ones(self.state.train_dataset.get_example_count()))
            test_data = self._get_dataclass_object_from_dataset(self.state.test_dataset,
                                                                np.zeros(self.state.test_dataset.get_example_count()),
                                                                np.zeros(self.state.test_dataset.get_example_count()))

            combined_data = merge_dataclass_objects([org_data, test_data, gen_data], fill_unmatched=True)
        else:
            org_data = self._get_dataclass_object_from_dataset(self.dataset, np.zeros(self.dataset.get_example_count()),
                                                               np.ones(self.dataset.get_example_count()))
            combined_data = merge_dataclass_objects([org_data, gen_data], fill_unmatched=True)

        bnp_write_to_file(path / f'combined_{self.state.name}_dataset.tsv', combined_data)

        metadata_yaml = SequenceDataset.create_metadata_dict(dataset_class=SequenceDataset.__name__,
                                             filename=f'combined_{self.state.name}_dataset.tsv',
                                             type_dict=type(combined_data).get_field_type_dict(all_fields=False),
                                             name=f'combined_{self.state.name}_dataset',
                                             labels={'gen_model_name': [self.method.name, ''], "from_gen_model": [True, False]})

        write_dataset_yaml(path / f'combined_{self.state.name}_dataset.yaml', metadata_yaml)

        self.state.combined_dataset = SequenceDataset.build(
            metadata_filename=path / f'combined_{self.state.name}_dataset.yaml',
            filename=path / f'combined_{self.state.name}_dataset.tsv', name=f'combined_{self.state.name}_dataset')

    def _get_dataclass_object_from_dataset(self, dataset: Dataset, from_gen_model_vals: np.ndarray,
                                           used_for_training_vals: np.ndarray):
        return dataset.data.add_fields(
            {'from_gen_model': np.where(from_gen_model_vals, 'T', 'F'),
             'used_for_training': np.where(used_for_training_vals, 'T', 'F')},
            {'from_gen_model': str, 'used_for_training': str})

    def _make_and_export_combined_dataset(self):
        if self.export_combined_dataset and isinstance(self.dataset, SequenceDataset):
            self._make_combined_dataset()
            export_path = PathBuilder.build(self.state.result_path / 'exported_combined_dataset')
            try:
                AIRRExporter.export(self.state.combined_dataset, export_path)
                self.state.exported_datasets['combined_dataset'] = export_path
            except AssertionError as e:
                logging.warning(f"{TrainGenModelInstruction.__name__}: {self.state.name}: combined dataset could not "
                                f"be exported due to the following error: {e}.")
        elif self.export_combined_dataset:
            logging.warning(f"{TrainGenModelInstruction.__name__}: {self.state.name}: export_combined_dataset is only "
                            f"supported for sequence datasets at this point.")

    def _run_reports(self):
        super()._run_reports()

        report_path = self._get_reports_path()
        for report in self.reports:
            original_dataset = self.state.train_dataset if self.state.training_percentage != 1 else self.dataset
            report.result_path = report_path
            if isinstance(report, TrainGenModelReport):
                report.generated_dataset = self.generated_dataset
                report.original_dataset = original_dataset
                report.model = self.method
                self.state.report_results['instruction_reports'].append(report.generate_report())
            elif isinstance(report, DataReport):
                rep = copy.deepcopy(report)
                rep.result_path = PathBuilder.build(rep.result_path.parent / f"{rep.result_path.name}_original_dataset")
                rep.dataset = original_dataset
                rep.name = rep.name + " (original dataset)"
                self.state.report_results['data_reports'].append(rep.generate_report())

        super()._print_report_summary_log()

    def _save_model(self):
        self.state.model_path = self.method.save_model(self.state.result_path / 'trained_model/')



import logging
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import plotly.express as px

from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class RepertoireClonotypeSummary(DataReport):
    """
    Shows the number of distinct clonotypes per repertoire in a given dataset as a bar plot.

    **Specification arguments:**

    - label (str): name of the label to use to color the plot, e.g., could be disease label, or None
    - split_by_label (bool): if True, the plot will be colored by the label specified in the label argument. Default is False.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_clonotype_summary_rep:
                    RepertoireClonotypeSummary:
                        label: celiac
                        split_by_label: true

    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None, number_of_processes: int = 1,
                 split_by_label: bool = None, label: str = None):
        super().__init__(dataset, result_path, name, number_of_processes)
        self.split_by_label = split_by_label
        self.label_name = label

    def _set_label_name(self):
        if self.split_by_label:
            if self.label_name is None:
                self.label_name = list(self.dataset.get_label_names())[0]
        else:
            self.label_name = None

    @classmethod
    def build_object(cls, **kwargs):
        ReportUtil.update_split_by_label_kwargs(kwargs, RepertoireClonotypeSummary.__name__)

        return RepertoireClonotypeSummary(**kwargs)

    def _generate(self) -> ReportResult:
        self._set_label_name()
        PathBuilder.build(self.result_path)
        return self._safe_plot()

    def _plot(self) -> ReportResult:
        clonotypes = pd.DataFrame.from_records(sorted([self._get_clonotype_count_with_label(repertoire) for repertoire in self.dataset.get_data()],
                                                      key=lambda x: x[0]), columns=['clonotype_count', self.label_name])
        clonotypes['repertoire'] = list(range(1, self.dataset.get_example_count()+1))
        clonotypes['repertoire_id'] = self.dataset.get_example_ids()
        fig = px.bar(clonotypes, x='repertoire', y='clonotype_count', color=self.label_name,
                     title='Clonotype count per repertoire',
                     color_discrete_sequence=px.colors.diverging.Tealrose)
        fig.update_layout(template="plotly_white", yaxis_title='clonotype count',
                          xaxis_title='repertoires sorted by clonotype count')
        clonotypes.to_csv(self.result_path / 'clonotype_count_per_repertoire.csv', index=False)
        fig.write_html(str(self.result_path / 'clonotype_count_per_repertoire.html'))

        return ReportResult(name=self.name, info="Clonotype count per repertoire",
                            output_figures=[ReportOutput(self.result_path / 'clonotype_count_per_repertoire.html')],
                            output_tables=[ReportOutput(self.result_path / 'clonotype_count_per_repertoire.csv')])

    def _get_clonotype_count_with_label(self, repertoire: Repertoire) -> Tuple[int, str]:

        sequences = repertoire.data.topandas()

        sequence_count = sequences.shape[0]
        unique_sequence_count = len(sequences.groupby(['cdr3_aa', 'v_call', 'j_call']).size().reset_index(name='count'))
        if sequence_count != unique_sequence_count:
            logging.warning(f"{RepertoireClonotypeSummary.__name__}: {self.name}: for repertoire {repertoire.identifier}, "
                            f"there are {sequence_count} sequences, but {unique_sequence_count} unique (CDR3 amino acid"
                            f" sequence, V call, J call) combinations.")

        return unique_sequence_count, repertoire.metadata[self.label_name] if self.split_by_label else None

    def check_prerequisites(self) -> bool:
        if isinstance(self.dataset, RepertoireDataset):
            return True
        else:
            logging.warning(f"{RepertoireClonotypeSummary.__name__}: report can be generated only from "
                            f"RepertoireDataset. Skipping this report...")
            return False



from dataclasses import dataclass
from pathlib import Path


@dataclass
class ReportOutput:
    def __init__(self, path: Path, name: str = None):
        self.path = Path(path)
        self.name = name

    def __repr__(self):
        return f"ReportOutput(name={self.name}, path={self.path})"



from functools import partial
from pathlib import Path
import bionumpy as bnp
from .TrainGenModelReport import TrainGenModelReport
from ..ReportOutput import ReportOutput
from ..ReportResult import ReportResult
from ...data_model import bnp_util
from ...data_model.SequenceParams import RegionType
from ...data_model.datasets.Dataset import Dataset
from ...environment.SequenceType import SequenceType
from ...ml_methods.generative_models.GenerativeModel import GenerativeModel
from ...ml_methods.generative_models.KLEvaluator import evaluate_similarities, KLEvaluator
from ...ml_methods.generative_models.MultinomialKmerModel import estimate_kmer_model
from ...util.PathBuilder import PathBuilder


class KLKmerComparison(TrainGenModelReport):
    """
    Estimates the KL divergence between the kmer-distributions of the original and generated dataset, and makes a plots
    that shows which sequences (and which kmers) contribute the most to the divergence.

    **Specification arguments:**

    - k (int): The kmer length to use for the KL divergence estimation. By default, k is set to 3.

    - n_sequences (int): The number of sequences to make the plot from (the sequences that contribute the most to the
      KL divergence). By default, n_sequences is set to 50.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        my_kl_report:
          KLKmerComparison:
            k: 3
            n_sequences: 50

    """

    def __init__(self, original_dataset: Dataset = None, generated_dataset: Dataset = None, result_path: Path = None,
                 name: str = None, number_of_processes: int = 1, model: GenerativeModel = None, k: int = 3,
                 n_sequences: int = 50, sequence_type: SequenceType = SequenceType.AMINO_ACID,
                 region_type: RegionType = RegionType.IMGT_CDR3):
        """
        The arguments defined below are set at runtime by the instruction.

        Args:

            original_dataset (Dataset): a dataset object (can be repertoire, receptor or sequence dataset, depending
            on the specific report) provided as input to the TrainGenModel instruction

            generated_dataset (Dataset): a dataset object as produced from the generative model after being trained on
            the original dataset

            result_path (Path): location where the results (plots, tables, etc.) will be stored

            name (str): user-defined name of the report used in the HTML overview automatically generated by the
            platform from the key used to define the report in the YAML

            number_of_processes (int): how many processes should be created at once to speed up the analysis.
            For personal machines, 4 or 8 is usually a good choice.

            model (GenerativeModel): trained generative model from the instruction

        """
        super().__init__(name=name, number_of_processes=number_of_processes)
        self.original_dataset = original_dataset
        self.generated_dataset = generated_dataset
        self.model = model
        self.result_path = result_path
        self.k = k
        self.n_sequences = n_sequences
        self.region_type = region_type
        self.sequence_type = sequence_type

    @staticmethod
    def get_title():
        return "KLTrainGenModel reports"

    @classmethod
    def build_object(cls, **kwargs):
        """
        Creates the object of the subclass of the Report class from the parameters so that it can be used in the analysis. Depending on the type of
        the report, the parameters provided here will be provided in parsing time, while the other necessary parameters (e.g., subset of the data from
        which the report should be created) will be provided at runtime. For more details, see specific direct subclasses of this class, describing
        different types of reports.

        Args:

            **kwargs: keyword arguments that will be provided by users in the specification (if immuneML is used as a command line tool) or in the
             dictionary when calling the method from the code, and which should be used to create the report object

        Returns:

            the object of the appropriate report class

        """
        location = cls.__name__
        return cls(**kwargs)

    def _compute_kl_divergence(self):
        """
        Computes the KL divergence between the original and generated dataset

        Returns:
            KL divergence value
        """
        return self._get_kmer_kl_evaluator()

    def _get_kmer_kl_evaluator(self):
        o_kmers, g_kmers = (bnp.sequence.get_kmers(getattr(dataset.data,
                                                           bnp_util.get_sequence_field_name(self.region_type,
                                                                                            self.sequence_type)),
                                                   self.k)
                            for dataset in (self.original_dataset, self.generated_dataset))
        estimator = partial(estimate_kmer_model, prior_count=1.0)
        return KLEvaluator(o_kmers, g_kmers, estimator, n_sequences=self.n_sequences)

    def _generate(self) -> ReportResult:
        """
        The function that needs to be implemented by the Report subclasses which actually creates the report (figures, tables, text files), depending
        on the specific aim of the report. After checking all prerequisites (e.g., if all parameters were set properly), generate_report() will call
        this function and return its result.

        Returns:

            ReportResult object which encapsulates all outputs (figure, table, and text files) so that they can be conveniently linked to in the
            final output of instructions

        """
        PathBuilder.build(self.result_path)
        evaluator = self._get_kmer_kl_evaluator()
        tables = [self._write_output_table(evaluator.get_worst_true_sequences(),
                                           self.result_path / "worst_true_sequences.tsv",
                                           name="Original sequences that don't fit the generated model"),
                  self._write_output_table(evaluator.get_worst_simulated_sequences(),
                                           self.result_path / "worst_simulated_sequences.tsv",
                                           name="Generated sequences that don't fit with the original model")]
        figures = []
        figures.append(self._plot_simulated(evaluator=evaluator))
        figures.append(self._plot_original(evaluator=evaluator))

        info_text = '''Estimated KL divergence between the kmer distributions in the original and generated datasets. Toghether with the sequences that contribute the most to the divergence.
        KL(original || generated) = {:.2f},  KL(generated || original) = {:.2f}'''.format(evaluator.true_kl(),
                                                                                          evaluator.simulated_kl())
        return ReportResult(name=self.name,
                            info=info_text,
                            output_figures=figures,
                            output_tables=[table for table in tables if table is not None])

    def _plot_simulated(self, evaluator: KLEvaluator):
        file_path = self.result_path / "bad_simulated_sequences.html"
        figure = evaluator.simulated_plot()
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Generated Sequences that contributes most to the KL divergence")

    def _plot_original(self, evaluator: KLEvaluator):
        file_path = self.result_path / "bad_original_sequences.html"
        figure = evaluator.original_plot()
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Original Sequences that contributes most to the KL divergence")



from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px

from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.BackgroundSequences import BackgroundSequences
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult


def report_signal_frequencies(frequencies: pd.DataFrame, path: Path) -> ReportResult:
    csv_output = ReportOutput(path / 'signal_frequencies.csv', 'signal frequencies as csv')
    frequencies.to_csv(csv_output.path, index=False)

    fig = px.bar(frequencies, x='signal', y='frequency', template='plotly_white', color_discrete_sequence=px.colors.diverging.Tealrose)
    fig_output = ReportOutput(path / 'signal_frequencies.html', 'signal frequencies bar chart')
    fig.write_html(str(fig_output.path))

    return ReportResult('signal frequencies', output_figures=[fig_output], output_tables=[csv_output])


def report_signal_cooccurrences(unique_values: np.ndarray, counts: np.ndarray, path: Path) -> ReportResult:
    df = pd.DataFrame({"signal_count": unique_values, "sequence_count": counts})
    csv_output = ReportOutput(path / 'signal_counts_per_sequence.csv', 'number of signals occurring in a single sequence')
    df.to_csv(csv_output.path, index=False)

    fig = px.bar(df, x='signal_count', y='sequence_count', template='plotly_white', color_discrete_sequence=px.colors.diverging.Tealrose)
    fig.update_layout(xaxis_title_text='signal count', yaxis_title_text='sequence count', xaxis={"tickmode": 'linear', "tick0": 0, "dtick": 1})
    fig_output = ReportOutput(path / 'signal_counts_per_sequence.html', 'signal counts per sequence')
    fig.write_html(str(fig_output.path))

    return ReportResult('signal co-occurrences', output_figures=[fig_output], output_tables=[csv_output])


def report_signal_joint_probs(signal_matrix: np.ndarray, signal_names: list, path: Path) -> ReportResult:
    joint_probs = np.zeros(shape=(len(signal_names), len(signal_names)))
    for outer_index, outer_signal in enumerate(signal_names):
        for inner_index, inner_signal in enumerate(signal_names):
            if inner_index == outer_index:
                joint_probs[inner_index, outer_index] = 1.
            elif inner_index < outer_index:
                joint_probs[inner_index, outer_index] = joint_probs[outer_index, inner_index]
            else:
                joint_probs[inner_index, outer_index] = np.logical_and(signal_matrix[:, inner_index], signal_matrix[:, outer_index]).sum() / signal_matrix.shape[0]

    table_output = ReportOutput(path / 'joint_probabilities_of_signals_co-occurring.csv', "joint probabilities of signals co-occurring")
    df = pd.DataFrame(joint_probs, index=signal_names, columns=signal_names)
    df.to_csv(str(table_output.path))

    fig = px.imshow(df, text_auto=True, color_continuous_scale='Aggrnyl')
    fig_output = ReportOutput(path / 'joint_probabilities_of_signals_co-occurring.html', "joint probabilities of signals co-occurring")
    fig.write_html(str(fig_output.path))

    return ReportResult('joint probabilities of signals co-occurring', output_figures=[fig_output], output_tables=[table_output])


def report_signal_cond_probs(signal_matrix: np.ndarray, signal_names: list, path: Path) -> ReportResult:
    cond_probs = np.zeros(shape=(len(signal_names), len(signal_names)))

    for outer_index, outer_signal in enumerate(signal_names):
        for inner_index, inner_signal in enumerate(signal_names):
            if inner_index == outer_index:
                cond_probs[outer_index, inner_index] = 1.
            else:
                cond_probs[outer_index, inner_index] = np.logical_and(signal_matrix[:, inner_index], signal_matrix[:, outer_index]).sum() / signal_matrix[:, inner_index].sum()

    description = "conditional probabilities of one signal given another signal"

    table_output = ReportOutput(path / 'cond_probabilities_of_signals_co-occurring.csv', description)
    df = pd.DataFrame(cond_probs, index=signal_names, columns=signal_names)
    df.to_csv(str(table_output.path))

    hovertext = [[f'P({signal_names[outer_ind]} | {signal_names[inner_ind]}) = ' \
                  f'{round(df.values[outer_ind, inner_ind], 3) if df.values[outer_ind, inner_ind] != np.nan else np.nan}'
                  for inner_ind in range(len(signal_names))] for outer_ind in range(len(signal_names))]

    fig = px.imshow(df, color_continuous_scale='Aggrnyl')
    fig.update_traces(text=df.values.round(3), texttemplate="%{text}", customdata=hovertext,
                      hovertemplate="%{customdata}<extra></extra>")
    fig.update_layout(template='plotly_white', xaxis_title='', yaxis_title='')
    fig_output = ReportOutput(path / 'cond_probabilities_of_signals_co-occurring.html', description)
    fig.write_html(str(fig_output.path))

    return ReportResult(description, output_figures=[fig_output], output_tables=[table_output])


def report_p_gen_histogram(sequences: BackgroundSequences, p_gen_bin_count: int, path: Path) -> ReportResult:
    log_p_gens = np.log10(sequences.p_gen)
    signal_info = ["_".join(s for index, s in enumerate(sequences.get_signal_names()) if el[index] == 1) for el in sequences.get_signal_matrix()]
    signal_info = [s if s != "" else "no signal" for s in signal_info]
    p_gen_df = pd.DataFrame({'log_p_gen': log_p_gens, "signal": signal_info})
    csv_output = ReportOutput(path / 'log10_p_gens.csv', 'generation probabilities on log10 scale')
    p_gen_df.to_csv(csv_output.path, index=False)

    fig_all = px.histogram(p_gen_df, x='log_p_gen', nbins=p_gen_bin_count + 1, template='plotly_white',
                           color_discrete_sequence=px.colors.diverging.Tealrose, histnorm='probability density')
    fig_all.update_layout(xaxis_title_text="logarithm of generation probability")
    fig_output_all = ReportOutput(path / 'log10_p_gens.html', 'generation probabilities on log10 scale')
    fig_all.write_html(str(fig_output_all.path))

    fig_signal = px.histogram(p_gen_df, x='log_p_gen', nbins=p_gen_bin_count + 1, template='plotly_white', color='signal', opacity=0.7,
                              color_discrete_sequence=px.colors.diverging.Tealrose, histnorm='probability density')
    fig_signal.update_layout(xaxis_title_text="logarithm of generation probability")
    fig_output_signal = ReportOutput(path / 'log10_p_gens_per_signal.html', 'generation probabilities on log10 scale per signal')
    fig_signal.write_html(str(fig_output_signal.path))

    return ReportResult('generation probabilities on log10 scale', output_figures=[fig_output_all, fig_output_signal], output_tables=[csv_output])


def report_seq_len_dist(sequences: BackgroundSequences, sequence_type: SequenceType, path: Path) -> ReportResult:
    lengths = sequences.get_sequence(sequence_type).lengths

    len_df = pd.DataFrame({"length": lengths})
    csv_output = ReportOutput(path / 'sequence_lengths.csv', 'sequence lengths')
    len_df.to_csv(csv_output.path, index=False)

    fig = px.histogram(len_df, x='length', template='plotly_white', color_discrete_sequence=px.colors.diverging.Tealrose,
                       histnorm='probability density')
    fig_output = ReportOutput(path / 'sequence_length_hist.html', 'sequence length histogram')
    fig.write_html(str(fig_output.path))

    return ReportResult('sequence length distribution', output_tables=[csv_output], output_figures=[fig_output])



import copy
import logging
import random
from pathlib import Path

import numpy as np
import pandas as pd
import torch
import yaml

from immuneML.data_model.EncodedData import EncodedData
from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.pytorch_implementations.PyTorchLogisticRegression import PyTorchLogisticRegression
from immuneML.ml_methods.util.Util import Util
from immuneML.util.PathBuilder import PathBuilder


class AtchleyKmerMILClassifier(MLMethod):
    """
    A binary Repertoire classifier which uses the data encoded by :ref:`AtchleyKmer` encoder to predict the repertoire label.

    The original publication:
    Ostmeyer J, Christley S, Toby IT, Cowell LG. Biophysicochemical motifs in T cell receptor sequences distinguish repertoires from tumor-infiltrating
    lymphocytes and adjacent healthy tissue. Cancer Res. Published online January 1, 2019:canres.2292.2018. `doi:10.1158/0008-5472.CAN-18-2292
    <https://cancerres.aacrjournals.org/content/79/7/1671>`_ .

    **Specification arguments:**

    - iteration_count (int): max number of training iterations

    - threshold (float): loss threshold at which to stop training if reached

    - evaluate_at (int): log model performance every 'evaluate_at' iterations and store the model every 'evaluate_at' iterations if early stopping
      is used

    - use_early_stopping (bool): whether to use early stopping

    - learning_rate (float): learning rate for stochastic gradient descent

    - random_seed (int): random seed used

    - zero_abundance_weight_init (bool): whether to use 0 as initial weight for abundance  term (if not, a random value is sampled from normal
      distribution with mean 0 and variance 1 / total_number_of_features

    - number_of_threads: number of threads to be used for training

    - initialization_count (int): how many times to repeat the fitting procedure from the beginning before choosing the optimal model (trains the model with multiple random initializations)

    - pytorch_device_name (str): The name of the pytorch device to use. This name will be passed to torch.device(pytorch_device_name).

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_kmer_mil_classifier:
                    AtchleyKmerMILClassifier:
                        iteration_count: 100
                        evaluate_at: 15
                        use_early_stopping: False
                        learning_rate: 0.01
                        random_seed: 100
                        zero_abundance_weight_init: True
                        number_of_threads: 8
                        threshold: 0.00001
                        initialization_count: 4

    """

    MIN_SEED_VALUE = 1
    MAX_SEED_VALUE = 100000

    def __init__(self, iteration_count: int = None, threshold: float = None, evaluate_at: int = None, use_early_stopping: bool = None,
                 random_seed: int = None, learning_rate: float = None, zero_abundance_weight_init: bool = None, number_of_threads: int = None,
                 initialization_count: int = None, pytorch_device_name: str = None):
        super().__init__()
        self.logistic_regression = None
        self.random_seed = random_seed
        self.iteration_count = iteration_count
        self.threshold = threshold
        self.evaluate_at = evaluate_at
        self.use_early_stopping = use_early_stopping
        self.learning_rate = learning_rate
        self.zero_abundance_weight_init = zero_abundance_weight_init
        self.number_of_threads = number_of_threads
        self.input_size = 0
        self.initialization_count = initialization_count
        self.pytorch_device_name = pytorch_device_name

    def _make_log_reg(self):
        return PyTorchLogisticRegression(in_features=self.input_size, zero_abundance_weight_init=self.zero_abundance_weight_init)

    def _fit(self, encoded_data: EncodedData, cores_for_training: int):
        mapped_y = Util.map_to_new_class_values(encoded_data.labels[self.label.name], self.class_mapping)
        self.logistic_regression = None
        min_loss = np.inf

        for initialization in range(self.initialization_count):

            random.seed(self.random_seed)
            random_seed = random.randint(AtchleyKmerMILClassifier.MIN_SEED_VALUE, AtchleyKmerMILClassifier.MAX_SEED_VALUE)

            Util.setup_pytorch(self.number_of_threads, random_seed, self.pytorch_device_name)
            self.input_size = encoded_data.examples.shape[1]

            log_reg = self._make_log_reg()
            loss = np.inf

            state = {"loss": loss, "model": None}
            loss_func = torch.nn.BCEWithLogitsLoss(reduction='mean')
            optimizer = torch.optim.SGD(log_reg.parameters(), lr=self.learning_rate)

            for iteration in range(self.iteration_count):

                # reset gradients
                optimizer.zero_grad()

                # compute predictions only for k-mers with max score
                max_logit_indices = self._get_max_logits_indices(encoded_data.examples, log_reg)
                example_count = encoded_data.examples.shape[0]
                examples = torch.from_numpy(encoded_data.examples).float()[torch.arange(example_count).long(), :, max_logit_indices]
                logits = log_reg(examples)

                # compute the loss
                loss = loss_func(logits, torch.tensor(mapped_y).float())

                # perform update
                loss.backward()
                optimizer.step()

                # log current score and keep model for early stopping if specified
                if iteration % self.evaluate_at == 0 or iteration == self.iteration_count - 1:
                    logging.info(f"AtchleyKmerMILClassifier: log loss at iteration {iteration + 1}/{self.iteration_count}: {loss}.")
                    if state["loss"] < loss and self.use_early_stopping:
                        state = {"loss": loss.numpy(), "model": copy.deepcopy(log_reg)}

                if loss < self.threshold:
                    break

                logging.warning(f"AtchleyKmerMILClassifier: the logistic regression model did not converge.")

            if loss > state['loss'] and self.use_early_stopping:
                log_reg.load_state_dict(state["model"])

            if min_loss > loss:
                self.logistic_regression = log_reg
                min_loss = loss

    def _get_max_logits_indices(self, data, log_reg=None):
        with torch.no_grad():
            if log_reg:
                logits = log_reg(torch.from_numpy(np.swapaxes(data, 1, 2).reshape(data.shape[0] * data.shape[2], -1)))
            else:
                logits = self.logistic_regression(torch.from_numpy(np.swapaxes(data, 1, 2).reshape(data.shape[0] * data.shape[2], -1)))
        logits = torch.reshape(logits, (data.shape[0], data.shape[2]))
        max_logits_indices = torch.argmax(logits, dim=1)
        return max_logits_indices.long()

    def _predict(self, encoded_data: EncodedData):
        predictions_proba = self._predict_proba(encoded_data)
        return {self.label.name: [self.class_mapping[val] for val in (predictions_proba[self.label.name][self.label.positive_class] > 0.5).tolist()]}

    def store(self, path: Path):
        PathBuilder.build(path)
        torch.save(copy.deepcopy(self.logistic_regression).state_dict(), str(path / "log_reg.pt"))
        custom_vars = copy.deepcopy(vars(self))

        coefficients_df = pd.DataFrame(custom_vars["logistic_regression"].linear.weight.detach().numpy(), columns=self.feature_names)
        coefficients_df["bias"] = custom_vars["logistic_regression"].linear.bias.detach().numpy()
        coefficients_df.to_csv(path / "coefficients.csv", index=False)

        del custom_vars["result_path"]
        del custom_vars["logistic_regression"]
        del custom_vars["label"]

        if self.label:
            custom_vars["label"] = self.label.get_desc_for_storage()

        params_path = self._get_custom_params_path(path)
        with params_path.open('w') as file:
            yaml.dump(custom_vars, file)

    def load(self, path):
        params_path = self._get_custom_params_path(path)
        with params_path.open("r") as file:
            custom_params = yaml.load(file, Loader=yaml.SafeLoader)

        for param, value in custom_params.items():
            if hasattr(self, param):
                if param == "label":
                    setattr(self, "label", Label(**value))
                else:
                    setattr(self, param, value)

        self.logistic_regression = self._make_log_reg()
        self.logistic_regression.load_state_dict(torch.load(str(path / "log_reg.pt")))

    def _get_custom_params_path(self, path):
        return path / "custom_params.yaml"

    def get_params(self):
        params = copy.deepcopy(vars(self))
        params["logistic_regression"] = copy.deepcopy(self.logistic_regression).state_dict()
        return params

    def _predict_proba(self, encoded_data: EncodedData):
        self.logistic_regression.eval()
        example_count = encoded_data.examples.shape[0]
        max_logit_indices = self._get_max_logits_indices(encoded_data.examples)
        with torch.no_grad():
            data = torch.from_numpy(encoded_data.examples).float()[torch.arange(example_count).long(), :, max_logit_indices]
            predictions = torch.sigmoid(self.logistic_regression(data)).numpy()

        return {self.label.name: {self.label.positive_class: predictions,
                                  self.label.get_binary_negative_class(): 1 - predictions}}

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return False

    def get_compatible_encoders(self):
        from immuneML.encodings.atchley_kmer_encoding.AtchleyKmerEncoder import AtchleyKmerEncoder
        return [AtchleyKmerEncoder]



from sklearn.linear_model import LogisticRegression as SklearnLogisticRegression

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class LogisticRegression(SklearnMethod):
    """
    This is a wrapper of scikit-learn’s LogisticRegression class. Please see the
    `LogisticRegression scikit-learn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html>`_
    of LogisticRegression for the parameters.

    Note: if you are interested in plotting the coefficients of the logistic regression model,
    consider running the :ref:`Coefficients` report.

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_logistic_regression: # user-defined method name
                    LogisticRegression: # name of the ML method
                        # sklearn parameters (same names as in original sklearn class)
                        penalty: l1 # always use penalty l1
                        C: [0.01, 0.1, 1, 10, 100] # find the optimal value for C
                        # Additional parameter that determines whether to print convergence warnings
                        show_warnings: True
                    # if any of the parameters under LogisticRegression is a list and model_selection_cv is True,
                    # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                    # and the optimal model will be selected
                    model_selection_cv: True
                    model_selection_n_folds: 5
                # alternative way to define ML method with default values:
                my_default_logistic_regression: LogisticRegression

    """

    default_parameters = {"max_iter": 1000, "solver": "saga"}

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        parameters = {**self.default_parameters, **(parameters if parameters is not None else {})}

        if parameter_grid is not None:
            parameter_grid = parameter_grid
        else:
            parameter_grid = {"max_iter": [1000]}

        super(LogisticRegression, self).__init__(parameter_grid=parameter_grid, parameters=parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        params = self._parameters.copy()
        params["n_jobs"] = cores_for_training
        return SklearnLogisticRegression(**params)

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return True

    def get_params(self):
        params = self.model.get_params()
        params["coefficients"] = self.model.coef_[0].tolist()
        params["intercept"] = self.model.intercept_.tolist()
        return params

    @staticmethod
    def get_documentation():
        doc = str(LogisticRegression.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation(
                "LogisticRegression"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc



from dataclasses import dataclass

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.example_weighting.ExampleWeightingParams import ExampleWeightingParams
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.workflows.steps.StepParams import StepParams


@dataclass
class DataWeighterParams(StepParams):

    dataset: Dataset
    weighting_strategy: ExampleWeightingStrategy
    weighting_params: ExampleWeightingParams



import copy
import itertools
import logging
import shutil
import typing
from datetime import datetime
from abc import ABC
from dataclasses import dataclass, fields
from pathlib import Path
from typing import List
from uuid import uuid4

import numpy as np
import pandas as pd
from bionumpy import get_bufferclass_for_datatype

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.SequenceSet import Receptor, ReceptorSequence, AIRRSequenceSet, \
    build_dynamic_airr_sequence_set_dataclass, make_receptors_from_data, make_sequences_from_data
from immuneML.data_model.bnp_util import bnp_write_to_file, bnp_read_from_file, read_yaml, \
    extend_dataclass_with_dynamic_fields, write_dataset_yaml, make_full_airr_seq_set_df
from immuneML.data_model.datasets.Dataset import Dataset


@dataclass
class ElementDataset(Dataset, ABC):
    filename: Path = None
    element_count: int = None
    element_ids: list = None
    dataset_file: Path = None
    dynamic_fields: dict = None
    bnp_dataclass: typing.Type = None
    _buffer_type = None

    def __post_init__(self):
        if self.dynamic_fields is None and self.dataset_file is not None:
            metadata = read_yaml(self.dataset_file)
            self.dynamic_fields = {key: AIRRSequenceSet.STR_TO_TYPE[val]
                                   for key, val in metadata['type_dict_dynamic_fields'].items()}
        if self.bnp_dataclass is None and self.dynamic_fields is not None:
            self.bnp_dataclass = extend_dataclass_with_dynamic_fields(AIRRSequenceSet,
                                                                      list(self.dynamic_fields.items()))
        if self.identifier is None:
            self.identifier = uuid4().hex

    @classmethod
    def create_metadata_dict(cls, dataset_class, filename, type_dict, name, labels, identifier=None):
        return {"type_dict_dynamic_fields": {key: AIRRSequenceSet.TYPE_TO_STR[val] for key, val in type_dict.items()},
                "identifier": identifier if identifier is not None else uuid4().hex,
                "dataset_type": dataset_class if isinstance(dataset_class, str) else dataset_class.__name__,
                "filename": filename,
                "name": name,
                "labels": {} if labels is None else labels,
                "timestamp": str(datetime.now())}

    @property
    def buffer_type(self):
        if not self._buffer_type:
            self._buffer_type = get_bufferclass_for_datatype(self.bnp_dataclass, delimiter='\t', has_header=True)
        return self._buffer_type

    @property
    def data(self) -> AIRRSequenceSet:
        return bnp_read_from_file(self.filename, self.buffer_type, self.bnp_dataclass)

    def clone(self, keep_identifier: bool = False):
        dataset = self.__class__(labels=self.labels, encoded_data=copy.deepcopy(self.encoded_data),
                                 filename=self.filename, bnp_dataclass=self.bnp_dataclass,
                                 dataset_file=copy.deepcopy(self.dataset_file),
                                 name=self.name)
        if keep_identifier:
            dataset.identifier = self.identifier
        dataset.element_ids = self.element_ids
        return dataset

    def get_label_names(self):
        invalid_label_names = ['type_dict_dynamic_fields', 'organism', 'j_gene', 'v_gene', 'j_call', 'v_call']
        return [el for el in list(self.labels.keys()) if el not in invalid_label_names] \
            if isinstance(self.labels, dict) else []


class SequenceDataset(ElementDataset):
    """A dataset class for sequence datasets (single chain). """

    @classmethod
    def build(cls, filename: Path, metadata_filename: Path, name: str = None, bnp_dc=None, labels: dict = None):
        metadata = read_yaml(metadata_filename)

        dynamic_fields = {key: AIRRSequenceSet.STR_TO_TYPE[val]
                          for key, val in metadata['type_dict_dynamic_fields'].items()}
        if bnp_dc is None:
            bnp_dc = extend_dataclass_with_dynamic_fields(AIRRSequenceSet, list(dynamic_fields.items()))

        if labels is None and 'labels' in metadata:
            labels = metadata['labels']

        return SequenceDataset(name=name, filename=filename, dataset_file=metadata_filename,
                               dynamic_fields=dynamic_fields, labels=labels,
                               bnp_dataclass=bnp_dc)

    @classmethod
    def build_from_objects(cls, sequences: List[ReceptorSequence], path: Path, name: str = None,
                           labels: dict = None, region_type: RegionType = RegionType.IMGT_CDR3):
        name = name if name is not None else uuid4().hex
        filename = path / f"{name}.tsv"

        all_fields_dict = make_all_fields_dict_from_sequences(sequences, region_type)
        bnp_dc, type_dict = build_dynamic_airr_sequence_set_dataclass(all_fields_dict)
        dc_object = bnp_dc(**all_fields_dict)
        bnp_write_to_file(filename, dc_object)

        dataset_metadata = cls.create_metadata_dict(dataset_class=cls.__name__,
                                                    filename=filename,
                                                    type_dict=type_dict,
                                                    name=name,
                                                    labels=labels)

        metadata_filename = path / f'{name}.yaml'
        write_dataset_yaml(metadata_filename, dataset_metadata)

        return SequenceDataset(filename=filename, name=name, labels=labels, dynamic_fields=type_dict,
                               dataset_file=metadata_filename, bnp_dataclass=bnp_dc,
                               identifier=dataset_metadata["identifier"])

    @classmethod
    def build_from_partial_df(cls, df: pd.DataFrame, path: Path, name: str = None, labels: dict = None,
                              type_dict: dict = None):

        airr_df = make_full_airr_seq_set_df(df)
        name = name if name is not None else uuid4().hex
        filename = path / f"{name}.tsv"

        airr_df.to_csv(filename, sep='\t', index=False)

        dataset_yaml = SequenceDataset.create_metadata_dict(SequenceDataset, filename=filename, type_dict=type_dict,
                                                            name=name, labels=labels)

        write_dataset_yaml(path / f'{name}.yaml', dataset_yaml)

        return SequenceDataset.build(filename, path / f'{name}.yaml', name)

    def get_metadata(self, field_names: list, return_df: bool = False):
        """Returns a dict or an equivalent pandas DataFrame with metadata information from ReceptorSequence objects for
        provided field names"""
        result = self.data.topandas()[field_names]
        result = fix_empty_strings_in_metadata(result)

        return result if return_df else result.to_dict("list")

    def get_attribute(self, attribute_name):
        return getattr(self.data, attribute_name)

    def make_subset(self, example_indices, path, dataset_type: str):
        data = self.data[example_indices]
        name = f"subset_{self.name}_{dataset_type}"
        data_filename = path / f'{name}.tsv'

        bnp_write_to_file(data_filename, data)

        metadata_filename = path / f'{name}.yaml'
        shutil.copyfile(self.dataset_file, metadata_filename)

        return SequenceDataset(filename=data_filename, name=name, labels=copy.deepcopy(self.labels),
                               dynamic_fields=self.dynamic_fields, dataset_file=metadata_filename,
                               bnp_dataclass=self.bnp_dataclass)

    def get_example_count(self):
        if self.element_count is None:
            self.element_count = len(self.data)
        return self.element_count

    def get_data(self, batch_size: int = 1, region_type: RegionType = RegionType.IMGT_CDR3):
        return make_sequences_from_data(self.data, self.dynamic_fields, region_type)

    def get_example_ids(self):
        return self.data.sequence_id.tolist()

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.data[start_index: end_index]


class ReceptorDataset(ElementDataset):
    """A dataset class for receptor datasets (paired chain)."""

    @classmethod
    def build(cls, filename: Path, metadata_filename: Path, name: str = None, bnp_dc=None, labels: dict = None):
        metadata = read_yaml(metadata_filename)
        dynamic_fields = {k: AIRRSequenceSet.STR_TO_TYPE[v] for k, v in metadata['type_dict_dynamic_fields'].items()}

        if bnp_dc is None:
            bnp_dc = extend_dataclass_with_dynamic_fields(AIRRSequenceSet, list(dynamic_fields.items()))

        if labels is None and 'labels' in metadata:
            labels = metadata['labels']

        return ReceptorDataset(name=name, filename=filename, dataset_file=metadata_filename,
                               dynamic_fields=dynamic_fields, labels=labels,
                               bnp_dataclass=bnp_dc)

    @classmethod
    def build_from_objects(cls, receptors: List[Receptor], path: Path, name: str = None,
                           labels: dict = None, region_type: RegionType = RegionType.IMGT_CDR3):
        name = name if name is not None else uuid4().hex
        filename = path / f"{name}.tsv"

        all_fields_dict = make_all_fields_dict_from_receptors(receptors, region_type)
        bnp_dc, type_dict = build_dynamic_airr_sequence_set_dataclass(all_fields_dict)
        bnp_write_to_file(filename, bnp_dc(**all_fields_dict))

        metadata_filename = path / f'{name}.yaml'
        dataset_metadata = cls.create_metadata_dict(dataset_class=cls.__name__,
                                                    filename=filename,
                                                    type_dict=type_dict,
                                                    name=name,
                                                    labels=labels)

        write_dataset_yaml(metadata_filename, dataset_metadata)

        return ReceptorDataset(filename=filename, name=name, labels=labels, dynamic_fields=type_dict,
                               dataset_file=metadata_filename, bnp_dataclass=bnp_dc,
                               identifier=dataset_metadata['identifier'])

    def get_metadata(self, field_names: list, return_df: bool = False):
        """Returns a dict or an equivalent pandas DataFrame with metadata information from Receptor objects for
        provided field names"""
        result = fix_empty_strings_in_metadata(self.data.topandas())

        result = result.groupby('cell_id', sort=False).agg(
            lambda x: "_".join([str(el) for el in list(set(x))])).reset_index()
        if any(field_name not in result.columns for field_name in field_names):
            logging.warning(
                f"ReceptorDataset {self.name}: not all requested metadata fields are available: {field_names}.")
        else:
            result = result[field_names]

        return result if return_df else result.to_dict('list')

    def make_subset(self, example_indices, path, dataset_type: str):
        true_indices = np.array([[ind * 2, ind * 2 + 1] for ind in example_indices]).flatten()
        data = self.data[true_indices]
        name = f"{dataset_type}_subset_{self.name}"

        bnp_write_to_file(path / f"{name}.tsv", data)

        metadata_filename = path / f'{name}.yaml'
        metadata = read_yaml(self.dataset_file)
        write_dataset_yaml(metadata_filename, {
            **metadata, **{'filename': f"{name}.tsv", 'name': name}
        })

        return ReceptorDataset(filename=path / f"{name}.tsv", name=name,
                               labels=copy.deepcopy(self.labels),
                               dynamic_fields=self.dynamic_fields, dataset_file=metadata_filename,
                               bnp_dataclass=self.bnp_dataclass)

    def get_example_count(self):
        if self.element_count is None:
            self.element_count = len(self.data) // 2
        return self.element_count

    def get_data(self, batch_size: int = 1, region_type: RegionType = RegionType.IMGT_CDR3):
        return make_receptors_from_data(self.data, self.dynamic_fields,
                                        f"ReceptorDataset {self.identifier}", region_type)

    def get_example_ids(self):
        return np.unique(self.data.cell_id.tolist()).tolist()

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.data[start_index * 2: end_index * 2]


def make_all_fields_dict_from_receptors(receptors: List[Receptor], region_type: RegionType = RegionType.IMGT_CDR3):
    sequences = list(itertools.chain.from_iterable([r.chain_1, r.chain_2] for r in receptors))
    for index, receptor in enumerate(receptors):
        sequences[index * 2].metadata = {**sequences[index * 2].metadata, **receptor.metadata,
                                         'receptor_id': receptor.receptor_id, 'cell_id': receptor.cell_id}
        sequences[index * 2 + 1].metadata = {**sequences[index * 2 + 1].metadata, **receptor.metadata,
                                             'receptor_id': receptor.receptor_id, 'cell_id': receptor.cell_id}

    return make_all_fields_dict_from_sequences(sequences, region_type)


def make_all_fields_dict_from_sequences(sequences: List[ReceptorSequence],
                                        region_type: RegionType = RegionType.IMGT_CDR3):
    airr_fields = fields(AIRRSequenceSet)
    all_fields = {seq_field.name: [] for seq_field in airr_fields}
    dynamic_fields = {}

    for index, sequence in enumerate(sequences):

        for key in sequence.metadata.keys():
            if key in all_fields:
                all_fields[key].append(sequence.metadata[key])
            elif key in dynamic_fields:
                dynamic_fields[key].append(sequence.metadata[key])
            else:
                dynamic_fields[key] = ['' for _ in range(index)] + [sequence.metadata[key]]

        for missing_key in dynamic_fields.keys():
            if len(dynamic_fields[missing_key]) == index:
                dynamic_fields[missing_key].append('')

        keys_not_to_add = ['metadata', 'sequence_aa', 'sequence'] + list(sequence.metadata.keys())
        for key in [f.name for f in fields(ReceptorSequence) if f.name not in keys_not_to_add]:
            all_fields[key].append(getattr(sequence, key))

        all_fields[region_type.value].append(sequence.sequence)
        all_fields[region_type.value + "_aa"].append(sequence.sequence_aa)

    all_fields = fill_in_neutral_vals(all_fields, airr_fields, sequences)

    return {**all_fields, **dynamic_fields}


def fill_in_neutral_vals(all_fields, airr_fields, sequences):
    for f in airr_fields:
        neutral_val = AIRRSequenceSet.get_neutral_value(f.type)
        if len(all_fields[f.name]) == 0:
            all_fields[f.name] = [neutral_val for _ in range(len(sequences))]
        elif any(val is None for val in all_fields[f.name]):
            all_fields[f.name] = [val if val is not None else neutral_val for val in all_fields[f.name]]

    return all_fields


def fix_empty_strings_in_metadata(df: pd.DataFrame):
    for col, col_type in df.dtypes.to_dict().items():
        if col_type == object:
            df[col] = df[col].astype(str).replace('nan', '')
    return df



from pathlib import Path

from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.workflows.instructions.GenModelInstruction import GenModelInstruction
from immuneML.workflows.instructions.GenModelInstruction import GenModelState


class ApplyGenModelState(GenModelState):
    pass


class ApplyGenModelInstruction(GenModelInstruction):
    """

    ApplyGenModel instruction implements applying generative AIRR models on the sequence level.

    This instruction takes as input a trained model (trained in the :ref:`TrainGenModel` instruction)
    which will be used for generating data and the number of sequences to be generated.
    It can also produce reports of the applied model and reports of generated sequences.


    **Specification arguments:**

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the applied model

    - reports (list): list of report ids (defined under definitions/reports) to apply after generating
      gen_examples_count examples; these can be data reports (to be run on generated examples), ML reports (to be run
      on the fitted model)

    - ml_config_path (str): path to the trained model in zip format (as provided by TrainGenModel instruction)

    **YAML specification:**

    .. highlight:: yaml
    .. code-block:: yaml

        instructions:
            my_apply_gen_model_inst: # user-defined instruction name
                type: ApplyGenModel
                gen_examples_count: 100
                ml_config_path: ./config.zip
                reports: [data_rep1, ml_rep2]

    """
    def __init__(self, method: GenerativeModel = None, reports: list = None, result_path: Path = None,
                 name: str = None, gen_examples_count: int = None):
        super().__init__(ApplyGenModelState(result_path, name, gen_examples_count), method, reports)

    def run(self, result_path: Path) -> GenModelState:
        self._set_path(result_path)
        self._gen_data()
        self._export_generated_dataset()
        self._run_reports()

        return self.state

    def _run_reports(self):
        super()._run_reports()
        super()._print_report_summary_log()



import copy
import subprocess
from multiprocessing.pool import Pool
from pathlib import Path
from typing import List

import math
import numpy as np
import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceParams import RegionType
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.abundance_encoding.AbundanceEncoderHelper import AbundanceEncoderHelper
from immuneML.encodings.abundance_encoding.CompAIRRBatchIterator import CompAIRRBatchIterator
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.CompAIRRHelper import CompAIRRHelper
from immuneML.util.CompAIRRParams import CompAIRRParams
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class CompAIRRSequenceAbundanceEncoder(DatasetEncoder):
    """
    This encoder works similarly to the :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder`,
    but internally uses `CompAIRR <https://github.com/uio-bmi/compairr/>`_ to accelerate core computations.

    This encoder represents the repertoires as vectors where:

    - the first element corresponds to the number of label-associated clonotypes
    - the second element is the total number of unique clonotypes

    To determine what clonotypes (amino acid sequences with or without matching V/J genes) are label-associated, Fisher's exact test (one-sided)
    is used.

    The encoder also writes out files containing the contingency table used for fisher's exact test,
    the resulting p-values, and the significantly abundant sequences
    (use :py:obj:`~immuneML.reports.encoding_reports.RelevantSequenceExporter.RelevantSequenceExporter` to export these sequences in AIRR format).

    Reference: Emerson, Ryan O. et al.
    ‘Immunosequencing Identifies Signatures of Cytomegalovirus Exposure History and HLA-Mediated Effects on the T Cell Repertoire’.
    Nature Genetics 49, no. 5 (May 2017): 659–65. `doi.org/10.1038/ng.3822 <https://doi.org/10.1038/ng.3822>`_.

    Note: to use this encoder, it is necessary to explicitly define the positive class for the label when defining the label
    in the instruction. With positive class defined, it can then be determined which sequences are indicative of the positive class.
    See :ref:`Reproduction of the CMV status predictions study` for an example using :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder`.

    **Dataset type:**

    - RepertoireDatasets


    **Specification arguments:**

    - p_value_threshold (float): The p value threshold to be used by the statistical test.

    - compairr_path (Path): optional path to the CompAIRR executable. If not given, it is assumed that CompAIRR
      has been installed such that it can be called directly on the command line with the command 'compairr',
      or that it is located at /usr/local/bin/compairr.

    - ignore_genes (bool): Whether to ignore V and J gene information. If False, the V and J genes between two receptor chains
      have to match. If True, gene information is ignored. By default, ignore_genes is False.

    - sequence_batch_size (int): The number of sequences in a batch when comparing sequences across repertoires, typically 100s of thousands.
      This does not affect the results of the encoding, but may affect the speed and memory usage. The default value is 1.000.000

    - threads (int): The number of threads to use for parallelization. This does not affect the results of the encoding, only the speed.
      The default number of threads is 8.

    - keep_temporary_files (bool): whether to keep temporary files, including CompAIRR input, output and log files, and the sequence
      presence matrix. This may take a lot of storage space if the input dataset is large. By default, temporary files are not kept.


    **YAML specification:**

    .. code-block:: yaml

        definitions:
            encodings:
                my_sa_encoding:
                    CompAIRRSequenceAbundance:
                        compairr_path: optional/path/to/compairr
                        p_value_threshold: 0.05
                        ignore_genes: False
                        threads: 8

    """

    RELEVANT_SEQUENCE_ABUNDANCE = "relevant_sequence_abundance"
    TOTAL_SEQUENCE_ABUNDANCE = "total_sequence_abundance"
    OUTPUT_FILENAME = "compairr_out.tsv"
    LOG_FILENAME = "compairr_log.txt"

    def __init__(self, p_value_threshold: float, compairr_path: str, sequence_batch_size: int, ignore_genes: bool,
                 keep_temporary_files: bool, threads: int, name: str = None):
        super().__init__(name=name)
        self.p_value_threshold = p_value_threshold
        self.sequence_batch_size = sequence_batch_size
        self.keep_temporary_files = keep_temporary_files

        self.repertoires_filepath = None
        self.sequences_filepaths = None
        self.relevant_indices_path = None
        self.relevant_sequence_path = None
        self.contingency_table_path = None
        self.p_values_path = None
        self.context = None
        self.compairr_sequence_presence = None

        self.compairr_params = CompAIRRParams(compairr_path=Path(compairr_path),
                                              keep_compairr_input=True,
                                              differences=0,
                                              indels=False,
                                              ignore_counts=True,
                                              ignore_genes=ignore_genes,
                                              threads=threads,
                                              output_filename=None,
                                              log_filename=None,
                                              output_pairs=False,
                                              pairs_filename=None)

    @staticmethod
    def _prepare_parameters(p_value_threshold: float, compairr_path: str, sequence_batch_size: int, ignore_genes: bool, keep_temporary_files: bool,
                            threads: int,
                            name: str = None):
        ParameterValidator.assert_type_and_value(p_value_threshold, float, "CompAIRRSequenceAbundanceEncoder", "p_value_threshold", min_inclusive=0,
                                                 max_inclusive=1)
        ParameterValidator.assert_type_and_value(sequence_batch_size, int, "CompAIRRSequenceAbundanceEncoder", "sequence_batch_size", min_inclusive=1)
        ParameterValidator.assert_type_and_value(ignore_genes, bool, "CompAIRRSequenceAbundanceEncoder", "ignore_genes")
        ParameterValidator.assert_type_and_value(keep_temporary_files, bool, "CompAIRRSequenceAbundanceEncoder", "keep_temporary_files")
        ParameterValidator.assert_type_and_value(threads, int, "CompAIRRSequenceAbundanceEncoder", "threads", min_inclusive=1)

        compairr_path = CompAIRRHelper.determine_compairr_path(compairr_path)

        return {
            "p_value_threshold": p_value_threshold,
            "compairr_path": compairr_path,
            "sequence_batch_size": sequence_batch_size,
            "ignore_genes": ignore_genes,
            "keep_temporary_files": keep_temporary_files,
            "threads": threads,
            "name": name
        }

    @staticmethod
    def build_object(dataset, **params):
        assert isinstance(dataset, RepertoireDataset), \
            "CompAIRRSequenceAbundanceEncoder: this encoding only works on repertoire datasets."
        prepared_params = CompAIRRSequenceAbundanceEncoder._prepare_parameters(**params)
        return CompAIRRSequenceAbundanceEncoder(**prepared_params)

    def encode(self, dataset, params: EncoderParams):
        EncoderHelper.check_positive_class_labels(params.label_config, CompAIRRSequenceAbundanceEncoder.__name__)
        self.compairr_params.is_cdr3 = params.region_type == RegionType.IMGT_CDR3

        self.compairr_sequence_presence = self._prepare_sequence_presence_data(dataset, params)

        return self._encode_data(dataset, params)

    def _prepare_sequence_presence_data(self, dataset, params):
        full_dataset = EncoderHelper.get_current_dataset(dataset, self.context)

        full_sequence_set = self._get_full_sequence_set(full_dataset)
        compairr_sequence_presence = self._get_sequence_presence(full_dataset, full_sequence_set, params)

        return compairr_sequence_presence

    def _get_full_sequence_set(self, full_dataset):
        full_sequence_set = CacheHandler.memo_by_params(self._build_dataset_params(full_dataset),
                                                        lambda: self.get_sequence_set(full_dataset))

        return full_sequence_set

    def get_sequence_set(self, repertoire_dataset):
        attributes = self.get_relevant_sequence_attributes()
        sequence_set = set()
        rep_seq_set_func = AbundanceEncoderHelper.get_matching_func_for_repertoire(attributes)

        for repertoire in repertoire_dataset.get_data():
            sequence_set.update(rep_seq_set_func(repertoire))

        return np.array(list(sequence_set))

    def _get_sequence_presence(self, full_dataset, full_sequence_set, params):
        compairr_sequence_presence = CacheHandler.memo_by_params(
            self._build_sequence_presence_params(full_dataset, self.compairr_params),
            lambda: self._compute_sequence_presence_with_compairr(full_dataset, full_sequence_set, params))

        return compairr_sequence_presence

    def _build_sequence_presence_params(self, dataset, compairr_params):
        return (("dataset_identifier", dataset.identifier),
                ("repertoire_ids", tuple(dataset.get_repertoire_ids())),
                ("ignore_genes", compairr_params.ignore_genes),
                ("indels", compairr_params.indels),
                ("differences", compairr_params.differences),
                ("ignore_counts", compairr_params.ignore_counts))

    def _build_dataset_params(self, dataset):
        return (("dataset_identifier", dataset.identifier),
                ("repertoire_ids", tuple(dataset.get_repertoire_ids())),
                "sequence_attributes", tuple(self.get_relevant_sequence_attributes()))

    def _compute_sequence_presence_with_compairr(self, dataset, full_sequence_set, params):
        self._prepare_compairr_input_files(dataset, full_sequence_set, params.result_path, params)

        arguments = [(sequences_filepath, params.result_path) for sequences_filepath in self.sequences_filepaths]

        with Pool(params.pool_size) as pool:
            paths = pool.starmap(self._new_run_compairr_on_batch, arguments)

        if not self.keep_temporary_files:
            self._remove_temporary_files(params.result_path)

        return CompAIRRBatchIterator(paths, self.sequence_batch_size)

    def _prepare_compairr_input_files(self, dataset, full_sequence_set, result_path, params: EncoderParams):
        PathBuilder.build(result_path)

        if self.repertoires_filepath is None:
            self.repertoires_filepath = result_path / "compairr_repertoires.tsv"

        CompAIRRHelper.write_repertoire_file(dataset, self.repertoires_filepath, self.compairr_params)

        if self.sequences_filepaths is None:
            self.sequences_filepaths = []

            n_sequence_files = math.ceil(len(full_sequence_set) / self.sequence_batch_size)
            subset_start_index = 0
            subset_end_index = min(self.sequence_batch_size, len(full_sequence_set))

            for file_index in range(n_sequence_files):
                filename = result_path / f"compairr_sequences_batch{file_index}.tsv"
                self.sequences_filepaths.append(filename)
                sequence_subset = full_sequence_set[subset_start_index:subset_end_index]

                self.write_sequence_set_file(sequence_subset, filename, offset=subset_start_index,
                                             region_type=params.region_type)

                subset_start_index += self.sequence_batch_size
                subset_end_index = min(subset_end_index + self.sequence_batch_size, len(full_sequence_set))

    def _remove_temporary_files(self, result_path):
        self.repertoires_filepath.unlink()

        for file in self.sequences_filepaths:
            file.unlink()

    def write_sequence_set_file(self, sequence_set, filename, offset=0, region_type=RegionType.IMGT_JUNCTION):
        sequence_col = 'junction' if region_type == RegionType.IMGT_JUNCTION else 'cdr3'
        if EnvironmentSettings.get_sequence_type() == SequenceType.AMINO_ACID:
            sequence_col = f"{sequence_col}_aa"
        vj_header = "" if self.compairr_params.ignore_genes else "\tv_call\tj_call"

        with open(filename, "w") as file:
            file.write(f"{sequence_col}{vj_header}\tduplicate_count\trepertoire_id\n")

            for id, sequence_info in enumerate(sequence_set, offset):
                file.write("\t".join(sequence_info) + f"\t1\t{id}\n")

    def _run_compairr_on_batch(self, sequences_filepath, result_path):
        batch = sequences_filepath.stem.split("_")[-1]
        compairr_params = copy.copy(self.compairr_params)
        compairr_params.output_filename = f"compairr_out_{batch}.txt"
        compairr_params.log_filename = f"compairr_log_{batch}.txt"

        args = CompAIRRHelper.get_cmd_args(compairr_params, [sequences_filepath, self.repertoires_filepath], result_path)
        compairr_result = subprocess.run(args, capture_output=True, text=True)
        return CompAIRRHelper.process_compairr_output_file(compairr_result, compairr_params, result_path)

    def _new_run_compairr_on_batch(self, sequences_filepath, result_path):
        batch = sequences_filepath.stem.split("_")[-1]
        compairr_params = copy.copy(self.compairr_params)
        compairr_params.output_filename = f"compairr_out_{batch}.txt"
        compairr_params.log_filename = f"compairr_log_{batch}.txt"

        args = CompAIRRHelper.get_cmd_args(compairr_params, [sequences_filepath, self.repertoires_filepath], result_path)
        compairr_result = subprocess.run(args, capture_output=True, text=True)
        return CompAIRRHelper.verify_compairr_output_path(compairr_result, compairr_params, result_path)

    def _encode_data(self, dataset: RepertoireDataset, params: EncoderParams):
        label = params.label_config.get_label_objects()[0]

        examples = self._calculate_abundance_matrix(dataset, self.compairr_sequence_presence, params)

        encoded_data = EncodedData(examples, dataset.get_metadata([label.name]) if params.encode_labels else None,
                                   dataset.get_repertoire_ids(),
                                   [CompAIRRSequenceAbundanceEncoder.RELEVANT_SEQUENCE_ABUNDANCE,
                                    CompAIRRSequenceAbundanceEncoder.TOTAL_SEQUENCE_ABUNDANCE],
                                   encoding=CompAIRRSequenceAbundanceEncoder.__name__,
                                   info={"relevant_sequence_path": self.relevant_sequence_path,
                                         "contingency_table_path": self.contingency_table_path,
                                         "p_values_path": self.p_values_path})

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _calculate_abundance_matrix(self, dataset: RepertoireDataset, compairr_sequence_presence, params: EncoderParams):
        repertoire_ids = dataset.get_repertoire_ids()
        compairr_sequence_presence.set_repertoire_ids(repertoire_ids)

        is_positive_class = AbundanceEncoderHelper.check_is_positive_class(dataset, repertoire_ids, params.label_config)

        relevant_sequence_indices, file_paths = AbundanceEncoderHelper\
            .get_relevant_sequence_indices(compairr_sequence_presence, is_positive_class, self.p_value_threshold,
                                           self.relevant_indices_path, params,
                                           cache_params=self._build_sequence_presence_params(dataset,
                                                                                             self.compairr_params))
        self._write_relevant_sequences_csv(dataset, relevant_sequence_indices, params.result_path)
        self._set_file_paths(file_paths)

        abundance_matrix = self._build_abundance_matrix(compairr_sequence_presence, repertoire_ids,
                                                        relevant_sequence_indices)

        return abundance_matrix

    def _build_abundance_matrix(self, compairr_sequence_presence, repertoire_ids, sequence_p_values_indices):
        abundance_matrix = np.zeros((len(repertoire_ids), 2))

        batch_start = 0

        for batch in compairr_sequence_presence.get_batches(repertoire_ids):
            batch_end = min(batch_start + self.sequence_batch_size, len(sequence_p_values_indices))

            partial_sequence_p_values_indices = sequence_p_values_indices[batch_start:batch_end]

            for idx, repertoire_id in enumerate(repertoire_ids):
                partial_repertoire_vector = batch[repertoire_id].to_numpy()

                relevant_sequence_abundance = np.sum(
                    partial_repertoire_vector[np.logical_and(partial_sequence_p_values_indices,
                                                             partial_repertoire_vector)])
                total_sequence_abundance = np.sum(partial_repertoire_vector)
                abundance_matrix[idx] += [relevant_sequence_abundance, total_sequence_abundance]

            batch_start = batch_start + self.sequence_batch_size

        return abundance_matrix

    def _set_file_paths(self, file_paths):
        self.relevant_indices_path = file_paths["relevant_indices_path"]
        self.contingency_table_path = file_paths["contingency_table_path"] if "contingency_table_path" in file_paths else None
        self.p_values_path = file_paths["p_values_path"] if "p_values_path" in file_paths else None

    def _write_relevant_sequences_csv(self, dataset, relevant_sequence_indices, result_path):
        full_dataset = EncoderHelper.get_current_dataset(dataset, self.context)
        full_sequence_set = self._get_full_sequence_set(full_dataset)
        relevant_sequences = full_sequence_set[relevant_sequence_indices]

        if self.relevant_sequence_path is None:
            self.relevant_sequence_path = result_path / 'relevant_sequences.csv'

        df = pd.DataFrame(relevant_sequences, columns=self.get_relevant_sequence_attributes())
        df.to_csv(self.relevant_sequence_path, sep=",", index=False)

    def get_relevant_sequence_attributes(self):
        attributes = ['cdr3_aa']

        if not self.compairr_params.ignore_genes:
            attributes += ["v_call", "j_call"]

        return attributes

    def set_context(self, context: dict):
        self.context = context
        return self

    def get_additional_files(self) -> List[Path]:
        return [file for file in [self.relevant_indices_path, self.relevant_sequence_path, self.contingency_table_path, self.p_values_path] if file]

    @staticmethod
    def load_encoder(encoder_file: Path):
        encoder = DatasetEncoder.load_encoder(encoder_file)
        encoder.relevant_indices_path = DatasetEncoder.load_attribute(encoder, encoder_file, "relevant_indices_path")

        return encoder



import uuid
from dataclasses import fields
from datetime import datetime
from pathlib import Path

import random
import pandas as pd

from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceSet import ReceptorSequence, Repertoire
from immuneML.data_model.bnp_util import build_dynamic_bnp_dataclass_obj, make_full_airr_seq_set_df, \
    write_dataset_yaml, write_yaml
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.util.PathBuilder import PathBuilder


class RepertoireBuilder:
    """
    Helper class for tests: creates repertoires from a list of a list of sequences and stores them in the given path
    """

    @staticmethod
    def build(sequences: list, path: Path, labels: dict = None, seq_metadata: list = None, subject_ids: list = None,
              name: str = "d1"):

        if subject_ids is not None:
            assert len(subject_ids) == len(sequences)

        if seq_metadata is not None:
            assert len(sequences) == len(seq_metadata)
            for index, sequence_list in enumerate(sequences):
                assert len(sequence_list) == len(seq_metadata[index])

        assert not (
                path / "repertoires").is_dir(), f"RepertoireBuilder: attempted to store new repertoires at {path / 'repertoires'} but this folder already exists. " \
                                                f"Please remove this folder or specify a different path. "

        PathBuilder.build(path)
        rep_path = PathBuilder.build(path / "repertoires")

        repertoires = []
        if subject_ids is None:
            subject_ids = []

        for rep_index, sequence_list in enumerate(sequences):
            if len(subject_ids) < len(sequences):
                subject_ids.append("rep_" + str(rep_index))

            df = pd.DataFrame({
                'cdr3_aa': sequence_list,
                'cdr3': ['' for _ in range(len(sequence_list))],
                'sequence_id': [uuid.uuid4().hex for _ in range(len(sequence_list))],
                'productive': ['T' for _ in range(len(sequence_list))],
                'vj_in_frame': ['T' for _ in range(len(sequence_list))],
                'stop_codon': ['F' for _ in range(len(sequence_list))]
            })

            if seq_metadata is None:
                df['v_call'], df['j_call'], df['duplicate_count'], df['locus'] = "TRBV1-1*01", "TRBJ1-1*01", [random.randint(1, 10) for _ in sequence_list], "TRB"
            else:
                df = pd.concat([df, pd.DataFrame.from_records(seq_metadata[rep_index])], axis=1)

            df = make_full_airr_seq_set_df(df)

            df.to_csv(str(rep_path / f'rep_{rep_index}.tsv'), sep='\t', index=False)

            if labels is not None:
                metadata = {key: labels[key][rep_index] for key in labels.keys()}
            else:
                metadata = {}

            metadata['type_dict_dynamic_fields'] = {key: AIRRSequenceSet.TYPE_TO_STR[df[key].dtype]
                                                    for key in df.columns
                                                    if key not in AIRRSequenceSet.get_field_type_dict()}

            metadata = {**metadata, **{"subject_id": subject_ids[rep_index]}}
            write_yaml(rep_path / f"rep_{rep_index}.yaml", metadata)

            bnp_dc_obj, _ = build_dynamic_bnp_dataclass_obj(df.to_dict(orient='list'))

            repertoire = Repertoire(rep_path / f"rep_{rep_index}.tsv", rep_path / f"rep_{rep_index}.yaml", metadata,
                                    _bnp_dataclass=type(bnp_dc_obj), identifier=uuid.uuid4().hex)
            repertoires.append(repertoire)

        df = pd.DataFrame({**{"filename": [repertoire.data_filename for repertoire in repertoires],
                              "subject_id": subject_ids,
                              "identifier": [repertoire.identifier for repertoire in repertoires]},
                           **(labels if labels is not None else {})})
        df.to_csv(path / f"{name}_metadata.csv", index=False)

        return repertoires, path / f"{name}_metadata.csv"

    @staticmethod
    def build_dataset(sequences: list, path: Path, labels: dict = None, seq_metadata: list = None,
                      subject_ids: list = None, name: str = "d1"):
        reps, metadata_file = RepertoireBuilder.build(sequences, path, labels, seq_metadata, subject_ids, name)

        # type_dict = {k: v for tmp_dict in [rep.metadata['type_dict_dynamic_fields'] for rep in reps]
        #              for k, v in tmp_dict.items()}

        labels_unique = {k: list(set(v)) for k, v in labels.items()} if isinstance(labels, dict) else {}
        identifier = uuid.uuid4().hex

        metadata_yaml = RepertoireDataset.create_metadata_dict(labels=labels_unique,
                                                               identifier=identifier,
                                                               metadata_file=str(metadata_file.name),
                                                               name=name)

        write_dataset_yaml(path / f'{name}.yaml', metadata_yaml)

        return RepertoireDataset(repertoires=reps, metadata_file=metadata_file, name=name, labels=labels_unique,
                                 dataset_file=path / f'{name}.yaml', identifier=identifier)



import inspect

from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class ObjectParser:

    @staticmethod
    def get_class_name(specs, valid_class_names, class_name_ending, location, key):
        if isinstance(specs, str):
            class_name = specs
        elif isinstance(specs, dict):
            assert len(specs) == 1, \
                f"{location}: More than one parameter passed to {class_name_ending.lower()} under key {key}. " \
                f"Only one value can be specified here. Valid options are: {str(valid_class_names)[1:-1]}"
            class_name = list(specs.keys())[0]
        else:
            raise AssertionError(f"{location}: Incorrect specification under key {key}.\n"
                                 f"Please choose one of the following:\n"
                                 f"(1) To specify an analysis component without parameters, simply specify the component as a single value, e.g;\n"
                                 f"my_component: ComponentName\n"
                                 f"\n"
                                 f"(2) To specify an analysis component with parameter values, specify it as a key containing "
                                 f"key-value pairs, e.g;\n"
                                 f"my_component: ComponentName\n"
                                 f"  parameter: value\n"
                                 f"\nMake sure the indentation is correct. Valid options are: {str(valid_class_names)[1:-1]}. \n"
                                 f"For more information and details on how to specify parameters, please refer to the documentation.")
        return class_name

    @staticmethod
    def get_params(specs, class_name):
        params = {}
        if isinstance(specs, dict):
            params = specs[class_name]
        return params

    @staticmethod
    def get_class(specs, valid_class_names, class_name_ending, class_path, location, key):
        class_name = ObjectParser.get_class_name(specs, valid_class_names, class_name_ending, location, key)
        cls = ReflectionHandler.get_class_by_name(f"{class_name}{class_name_ending}", class_path)
        return cls

    @staticmethod
    def get_all_params(specs, class_path, short_class_name, key: str = None):
        default_params = DefaultParamsLoader.load(class_path, short_class_name)
        specified_params = ObjectParser.get_params(specs, short_class_name)
        params = {**default_params, **specified_params, "name": key}
        return params

    @staticmethod
    def parse_object(specs, valid_class_names: list, class_name_ending: str, class_path: str, location: str, key: str,
                     builder: bool = False, return_params_dict: bool = False):
        class_name = ObjectParser.get_class_name(specs, valid_class_names, class_name_ending, location, key)
        ParameterValidator.assert_in_valid_list(class_name, valid_class_names, location, key)

        cls = ReflectionHandler.get_class_by_name(f"{class_name}{class_name_ending}", class_path)
        params = ObjectParser.get_all_params(specs, class_path, class_name, key)

        try:
            if "name" not in inspect.signature(cls.__init__).parameters.keys():
                del params["name"]
            obj = cls.build_object(**params) if builder and hasattr(cls, "build_object") else cls(**params)
        except TypeError as err:
            raise AssertionError(f"{location}: invalid parameter {err.args[0]} when specifying parameters in {specs} "
                                 f"under key {key}. Valid parameter names are: "
                                 f"{[name for name in inspect.signature(cls.__init__).parameters.keys()]}").with_traceback(err.__traceback__)

        return (obj, {class_name: params}) if return_params_dict else obj



class Constants:

    VERSION = "3.0.11"

    # encoding constants
    FEATURE_DELIMITER = "-"

    # importing constants
    UNKNOWN = ''
    ALLELE_DELIMITER = "*"
    GENE_DELIMITER = "-"
    STOP_CODON = "*"
    CACHE_TYPE = "cache_type"
    CACHE_PATH = 'cache_path'
    COMMENT_SIGN = "#"
    NOT_COMPUTED = "not computed"
    SIGNAL_DELIMITER = "__"
    GAP_LETTER = "-"

    # plotting constants
    PLOTLY_BLACK = "#2A3F5E"
